# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import struct
from typing import Any, Dict

from elftools.common.exceptions import ELFError, ELFParseError
from elftools.elf.dynamic import DynamicSection
from elftools.elf.elffile import ELFFile
from elftools.elf.enums import ENUM_DT_FLAGS, ENUM_DT_FLAGS_1
from elftools.elf.sections import NoteSection
from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def supports_file(filetype) -> bool:
    return filetype == "ELF"


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    return extract_elf_info(filename)


_EI_OSABI_NAME = {
    "ELFOSABI_SYSV": "System V",
    "ELFOSABI_HPUX": "HP-UX",
    "ELFOSABI_NETBSD": "NetBSD",
    "ELFOSABI_LINUX": "Linux",
    "ELFOSABI_HURD": "GNU/Hurd",
    "ELFOSABI_SOLARIS": "Solaris",
    "ELFOSABI_AIX": "AIX",
    "ELFOSABI_IRIX": "IRIX",
    "ELFOSABI_FREEBSD": "FreeBSD",
    "ELFOSABI_TRU64": "TRUE64",
    "ELFOSABI_MODESTO": "Novell Modesto",
    "ELFOSABI_OPENBSD": "OpenBSD",
    "ELFOSABI_OPENVMS": "OpenVMS",
    "ELFOSABI_NSK": "HP Non-Stop Kernel",
    "ELFOSABI_AROS": "AROS",
    "ELFOSABI_FENIXOS": "Fenix OS",
    "ELFOSABI_CLOUD": "Nuxi CloudABI",
    "ELFOSABI_SORTIX": "Sortix",
    "ELFOSABI_ARM_AEABI": "ARM EABI",
    "ELFOSABI_ARM": "ARM",
    "ELFOSABI_CELL_LV2": "CellOS Lv-2",
    "ELFOSABI_STANDALONE": "Standalone",
}


def extract_elf_info(filename: str) -> object:
    with open(filename, "rb") as f:
        try:
            elf = ELFFile(f)
        except (OSError, ELFError):
            return {}

        # Don't assume OS is Linux, map e_ident EI_OSABI value to an OS name
        file_details: Dict[str, Any] = {"OS": ""}
        file_details["elfIdent"] = get_elf_ident_from_file_header(f, elf.little_endian)
        file_details["elfDependencies"] = []
        file_details["elfRpath"] = []
        file_details["elfRunpath"] = []
        file_details["elfSoname"] = []
        file_details["elfInterpreter"] = []
        file_details["elfDynamicFlags"] = []
        file_details["elfDynamicFlags1"] = []
        file_details["elfGnuRelro"] = False
        file_details["elfComment"] = []
        file_details["elfNote"] = []
        file_details["elfOsAbi"] = elf["e_ident"]["EI_OSABI"]
        file_details["elfHumanArch"] = elf.get_machine_arch()
        file_details["elfArchNumber"] = file_details["elfIdent"]["E_MACHINE"]
        file_details["elfArchitecture"] = elf["e_machine"]

        # Get a human readable name for the OS
        file_details["OS"] = _EI_OSABI_NAME.get(file_details["elfOsAbi"], "")
        try:
            for section in elf.iter_sections():
                if section.name == ".interp":
                    file_details["elfInterpreter"].append(section.data().rstrip(b"\x00").decode())
                if section.name == ".comment":
                    for v in section.data().rstrip(b"\x00").split(b"\x00"):
                        file_details["elfComment"].append(v.decode())
                if isinstance(section, NoteSection):
                    for note in section.iter_notes():
                        # Info on contents of NetBSD and PaX notes: https://www.netbsd.org/docs/kernel/elf-notes.html
                        # Heuristics used by Avast RetDec to identify compiler/OS: https://github.com/avast/retdec/commit/d55b541c26fb110381b2203dc7baa50928e3f473
                        note_info = {}
                        note_info["sectionName"] = section.name
                        note_info["name"] = note.n_name
                        note_info["type"] = note.n_type
                        if note.n_name == "GNU":
                            if note.n_type == "NT_GNU_ABI_TAG":
                                note_info["os"] = note.n_desc.abi_os
                                note_info["abi"] = (
                                    f"{note.n_desc.abi_major}.{note.n_desc.abi_minor}.{note.n_desc.abi_tiny}"
                                )
                            elif note.n_type in ("NT_GNU_BUILD_ID", "NT_GNU_GOLD_VERSION"):
                                note_info["desc"] = note.n_desc
                            else:
                                note_info["descdata"] = note.n_descdata.decode("unicode_escape")
                        else:
                            note_info["descdata"] = note.n_descdata.decode("unicode_escape")
                        file_details["elfNote"].append(note_info)
                if isinstance(section, DynamicSection):
                    for tag in section.iter_tags():
                        if tag.entry.d_tag == "DT_NEEDED":
                            # Shared libraries
                            file_details["elfDependencies"].append(tag.needed)
                        elif tag.entry.d_tag == "DT_RPATH":
                            # Library rpath
                            file_details["elfRpath"].append(tag.rpath)
                        elif tag.entry.d_tag == "DT_RUNPATH":
                            # Library runpath
                            file_details["elfRunpath"].append(tag.runpath)
                        elif tag.entry.d_tag == "DT_SONAME":
                            # Library soname (for linking)
                            file_details["elfSoname"].append(tag.soname)
                        elif tag.entry.d_tag == "DT_FLAGS":
                            # Dynamic Flags, DT_FLAGS
                            dt_flags_entry: Dict[str, Any] = {}
                            dt_flags_entry["value"] = hex(tag.entry.d_val)
                            # $ORIGIN processing is required
                            dt_flags_entry["DF_ORIGIN"] = bool(
                                tag.entry.d_val & ENUM_DT_FLAGS["DF_ORIGIN"]
                            )
                            # Perform complete relocation processing (part of Full RELRO)
                            dt_flags_entry["DF_BIND_NOW"] = bool(
                                tag.entry.d_val & ENUM_DT_FLAGS["DF_BIND_NOW"]
                            )
                            file_details["elfDynamicFlags"].append(dt_flags_entry)
                        elif tag.entry.d_tag == "DT_FLAGS_1":
                            # Dynamic Flags, DT_FLAGS_1 (custom entry first added by binutils)
                            dt_flags_1_entry: Dict[str, Any] = {}
                            dt_flags_1_entry["value"] = hex(tag.entry.d_val)
                            # Position-Independent Executable file
                            dt_flags_1_entry["DF_1_PIE"] = bool(
                                tag.entry.d_val & ENUM_DT_FLAGS_1["DF_1_PIE"]
                            )
                            # Perform complete relocation processing
                            dt_flags_1_entry["DF_1_NOW"] = bool(
                                tag.entry.d_val & ENUM_DT_FLAGS_1["DF_1_NOW"]
                            )
                            # $ORIGIN processing is required
                            dt_flags_1_entry["DF_1_ORIGIN"] = bool(
                                tag.entry.d_val & ENUM_DT_FLAGS_1["DF_1_ORIGIN"]
                            )
                            # Ignore the default library search path
                            dt_flags_1_entry["DF_1_NODEFLIB"] = bool(
                                tag.entry.d_val & ENUM_DT_FLAGS_1["DF_1_NODEFLIB"]
                            )
                            file_details["elfDynamicFlags1"].append(dt_flags_1_entry)
        except ELFParseError as e:
            logger.warning(f"Error while parsing ELF sections in {filename}: {repr(e)}")

        # Check for presence of special segments (e.g. PT_GNU_RELRO)
        for segment in elf.iter_segments():
            if segment["p_type"] == "PT_GNU_RELRO":
                file_details["elfGnuRelro"] = True

        if elf["e_type"] == "ET_EXEC":
            file_details["elfIsExe"] = True
        else:
            file_details["elfIsExe"] = False

        if elf["e_type"] == "ET_DYN":
            file_details["elfIsLib"] = True
        else:
            file_details["elfIsLib"] = False

        if elf["e_type"] == "ET_REL":
            file_details["elfIsRel"] = True
        else:
            file_details["elfIsRel"] = False

        if elf["e_type"] == "ET_CORE":
            file_details["elfIsCore"] = True
        else:
            file_details["elfIsCore"] = False

        return file_details


def get_elf_ident_from_file_header(f, little_endian) -> dict:
    # Details on ELF header information/format:
    # https://man7.org/linux/man-pages/man5/elf.5.html
    # https://en.wikipedia.org/wiki/Executable_and_Linkable_Format#File_header
    e_ident = {}

    # EI_CLASS (32 or 64-bit architecture)
    f.seek(4)
    e_ident["EI_CLASS"] = struct.unpack("B", f.read(1))[0]
    # EI_DATA (file encoding/endianness)
    f.seek(5)
    e_ident["EI_DATA"] = struct.unpack("B", f.read(1))[0]
    # EI_VERSION (version number of ELF specification)
    f.seek(6)
    e_ident["EI_VERSION"] = struct.unpack("B", f.read(1))[0]
    # EI_OSABI
    f.seek(7)
    e_ident["EI_OSABI"] = struct.unpack("B", f.read(1))[0]
    # EI_ABIVERSION
    f.seek(8)
    e_ident["EI_ABIVERSION"] = struct.unpack("B", f.read(1))[0]
    # E_MACHINE
    f.seek(18)
    isa_data = f.read(2)
    if little_endian:
        e_ident["E_MACHINE"] = struct.unpack("<H", isa_data)[0]
    else:
        e_ident["E_MACHINE"] = struct.unpack(">H", isa_data)[0]

    return e_ident



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import List, Optional

# pylint: disable=too-many-instance-attributes


@dataclass
class File:
    filePath: str
    description: str
    category: str
    capturedBy: str
    captureTime: str
    source: str
    methodOfAcquisition: Optional[List[str]] = None



# Copyright 2023 Lawrence Livermore Natioanl Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class ContextEntry:
    extractPaths: List[str]
    archive: Optional[str] = None
    installPrefix: Optional[str] = None
    omitUnrecognizedTypes: Optional[bool] = None
    includeFileExts: Optional[List[str]] = None
    excludeFileExts: Optional[List[str]] = None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pathlib

from surfactant.relationships._internal.posix_utils import posix_normpath


def test_posix_normpath():
    assert posix_normpath("") == pathlib.PurePosixPath(".")
    assert posix_normpath("..") == pathlib.PurePosixPath(".")
    assert posix_normpath("../") == pathlib.PurePosixPath(".")
    assert posix_normpath("../..") == pathlib.PurePosixPath(".")

    assert posix_normpath("ab/../xy") == pathlib.PurePosixPath("xy")
    assert posix_normpath("/a/b//..///c") == pathlib.PurePosixPath("/a/c")

    assert posix_normpath("//..") == pathlib.PurePosixPath("//")
    assert posix_normpath("//../") == pathlib.PurePosixPath("//")
    assert posix_normpath("//../a") == pathlib.PurePosixPath("//a")
    assert posix_normpath("//./") == pathlib.PurePosixPath("//")
    assert posix_normpath("//./a") == pathlib.PurePosixPath("//a")
    assert posix_normpath("//./a/../b") == pathlib.PurePosixPath("//b")

    assert posix_normpath("///") == pathlib.PurePosixPath("/")
    assert posix_normpath("///a") == pathlib.PurePosixPath("/a")
    assert posix_normpath("///a/../b") == pathlib.PurePosixPath("/b")
    assert posix_normpath("///a/../") == pathlib.PurePosixPath("/")

    assert posix_normpath("////a/../b") == pathlib.PurePosixPath("/b")



import subprocess
import sys

import click

from surfactant.configmanager import ConfigManager
from surfactant.plugin.manager import (
    call_init_hooks,
    find_plugin_by_name,
    get_plugin_manager,
    is_hook_implemented,
    print_plugins,
)


@click.command(name="list")
def plugin_list_cmd():
    """Lists plugins."""
    pm = get_plugin_manager()
    print_plugins(pm)

    # List disabled plugins
    config_manager = ConfigManager()
    section = "core"
    section_key = "disable_plugins"

    # Retrieve the current list of plugins that should be blocked
    current_blocked_plugins = config_manager.get(section, section_key, [])

    print("\nDISABLED PLUGINS")
    for disabled_plugin in current_blocked_plugins:
        print(f"\tname: {disabled_plugin}")


@click.command(name="enable")
@click.argument("plugin_names", nargs=-1)
def plugin_enable_cmd(plugin_names):
    """Enables one or more plugins."""
    if not plugin_names:
        raise click.UsageError("At least one plugin name must be specified.")
    section = "core"
    section_key = "disable_plugins"

    config_manager = ConfigManager()

    # Retrieve the current list of plugins that should be blocked
    current_blocked_plugins = config_manager.get(section, section_key, [])

    # Ensure current_blocked_plugins is a list
    if isinstance(current_blocked_plugins, str):
        current_blocked_plugins = [current_blocked_plugins]

    # Remove the plugin from the blocked list if present
    for plugin_name in plugin_names:
        if plugin_name in current_blocked_plugins:
            current_blocked_plugins.remove(plugin_name)

    # Update the configuration to remove plugins from being disabled
    config_manager.set(section, section_key, current_blocked_plugins)
    click.echo(f"Updated blocked plugins: {current_blocked_plugins}")


@click.command(name="disable")
@click.argument("plugin_names", nargs=-1)
def plugin_disable_cmd(plugin_names):
    """Disables one or more plugins."""
    if not plugin_names:
        raise click.UsageError("At least one plugin name must be specified.")
    section = "core"
    section_key = "disable_plugins"

    config_manager = ConfigManager()

    # Retrieve the current list of plugins that should be blocked
    current_blocked_plugins = config_manager.get(section, section_key, [])

    # Ensure current_blocked_plugins is a list
    if isinstance(current_blocked_plugins, str):
        current_blocked_plugins = [current_blocked_plugins]

    # Add the plugin to the blocked list if not already present
    for plugin_name in plugin_names:
        if plugin_name not in current_blocked_plugins:
            current_blocked_plugins.append(plugin_name)

    # Update the configuration to add plugins to be disabled
    if current_blocked_plugins:
        config_manager.set(section, section_key, current_blocked_plugins)
        click.echo(f"Updated blocked plugins: {current_blocked_plugins}")


@click.command(name="install")
@click.argument("plugin_name")
def plugin_install_cmd(plugin_name):
    """Installs a plugin."""
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "install", plugin_name])
        click.echo(f"Successfully installed {plugin_name}.")
    except subprocess.CalledProcessError as e:
        click.echo(f"Failed to install {plugin_name}: {e}", err=True)


@click.command(name="uninstall")
@click.argument("plugin_name")
def plugin_uninstall_cmd(plugin_name):
    """Uninstalls a plugin."""
    try:
        subprocess.check_call([sys.executable, "-m", "pip", "uninstall", "-y", plugin_name])
        click.echo(f"Successfully uninstalled {plugin_name}.")
    except subprocess.CalledProcessError as e:
        click.echo(f"Failed to uninstall {plugin_name}: {e}", err=True)


@click.command(name="update-db")
@click.argument("plugin_name", required=False)
@click.option(
    "--all",
    "update_all",
    is_flag=True,
    help="Update all plugins that implement the 'update_db' hook.",
)
def plugin_update_db_cmd(plugin_name, update_all):
    """Updates the database for a specified plugin or all plugins if --all is used."""
    pm = get_plugin_manager()
    call_init_hooks(pm, hook_filter=["update_db"], command_name="update-db")

    if update_all:
        # Update all plugins that implement the update_db hook
        for plugin in pm.get_plugins():
            if is_hook_implemented(pm, plugin, "update_db"):
                plugin_name = pm.get_name(plugin) or pm.get_canonical_name(plugin)
                click.echo(f"Updating {plugin_name} ...")
                update_result = plugin.update_db()
                if update_result:
                    click.echo(f"Update result for {plugin_name}: {update_result}")
                else:
                    click.echo(f"No update operation performed for {plugin_name}.")
    else:
        if not plugin_name:
            click.echo("Please specify a plugin name or use --all to update all plugins.", err=True)
            return

        plugin = find_plugin_by_name(pm, plugin_name)  # Get an instance of the plugin

        # Check if the plugin is registered
        if not plugin:
            click.echo(f"Plugin '{plugin_name}' not found.", err=True)
            return

        # Check if the plugin has implemented the update_db hook
        has_update_db_hook = is_hook_implemented(pm, plugin, "update_db")

        if not has_update_db_hook:
            click.echo(f"Plugin '{plugin_name}' does not implement the 'update_db' hook.", err=True)
            return

        # Call the update_db hook for the specified plugin
        plugin_name = pm.get_name(plugin) or pm.get_canonical_name(plugin)
        click.echo(f"Updating {plugin_name} ...")
        update_result = plugin.update_db()

        if update_result:
            click.echo(f"Update result for {plugin_name}: {update_result}")
        else:
            click.echo(f"No update operation performed for {plugin_name}.")



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import uuid
from dataclasses import dataclass, field
from typing import List, Optional

from ._file import File
from ._provenance import ObservationProvenance

# pylint: disable=too-many-instance-attributes


@dataclass
class Observation:
    UUID: str = field(default_factory=lambda: str(uuid.uuid4()))
    flag: Optional[str] = None
    CWEClass: Optional[str] = None
    targetEnvironmentOrDevice: Optional[str] = None
    potentialEffectOrImpact: Optional[str] = None
    CVE: Optional[str] = None
    CVSS: Optional[int] = None
    analystInfo: Optional[str] = None
    discovery: Optional[str] = None
    files: Optional[List[File]] = None
    toRecreate: Optional[str] = None
    mitigationSuggestions: Optional[str] = None
    confidenceLevel: Optional[str] = None  # enum: none, low, medium, high, critical
    provenance: Optional[List[ObservationProvenance]] = None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pathlib
import string
from typing import Optional

import surfactant.plugin


def check_motorola(current_line):
    current_line = current_line.strip()
    if len(current_line) < 1:
        return False
    if current_line[0] != "S" and current_line[0] != "s":
        return False
    for x in range(1, len(current_line)):
        if current_line[x] not in string.hexdigits:
            return False
    return True


def check_intel(current_line):
    current_line = current_line.strip()
    if len(current_line) < 1:
        return False
    if current_line[0] != ":":
        return False
    for x in range(1, len(current_line)):
        if current_line[x] not in string.hexdigits:
            return False
    return True


# extensions from:
# https://en.wikipedia.org/wiki/Intel_HEX
# - not included: all p00 to pff extensions
# https://en.wikipedia.org/wiki/SREC_(file_format)
hex_file_extensions = [
    ".hex",
    ".mcs",
    ".h86",
    ".hxl",
    ".hxh",
    ".obl",
    ".obh",
    ".ihex",
    ".ihe",
    ".ihx",
    ".a43",
    ".a90",
    ".s-record",
    ".srecord",
    ".s-rec",
    ".srec",
    ".s19",
    ".s28",
    ".s37",
    ".s",
    ".s1",
    ".s2",
    ".s3",
    ".sx",
    ".exo",
    ".mot",
    ".mxt",
]


@surfactant.plugin.hookimpl
def identify_file_type(filepath: str) -> Optional[str]:
    file_suffix = pathlib.Path(filepath).suffix.lower()
    # quick exit based on file extension
    if file_suffix not in hex_file_extensions:
        return None
    try:
        with open(filepath, "r") as f:
            percent_intel = 0
            percent_motorola = 0
            for _ in range(100):
                curr = f.readline()
                if not curr:
                    break
                if check_motorola(curr):
                    percent_motorola += 1
                elif check_intel(curr):
                    percent_intel += 1
            if percent_intel > percent_motorola:
                return "INTEL_HEX"
            if percent_motorola > percent_intel:
                return "MOTOROLA_SREC"
            return None

    except (FileNotFoundError, UnicodeDecodeError):
        return None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from typing import Union

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def supports_file(filetype: str) -> bool:
    return filetype in ("A.OUT little", "A.OUT big")


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    return extract_a_out_info(filetype, filename)


# Machine ID/types can be found in various OS source files (+ file magic patterns) - some are ambiguous
# https://github.com/file/file/tree/master/magic/Magdir
# https://github.com/openbsd/src/blob/master/sys/sys/exec.h#L259-L291
# https://github.com/haiku/haiku/blob/master/src/tools/elf2aout.c#L70-L111
# https://github.com/KimLoanSA/mimuw/blob/main/sem4/so/abrams/zadania/zad6/minix_source/usr/include/sys/aout_mids.h#L36-L67
_A_OUT_TARGET_NAME = {
    0x00: "Unknown",
    0x01: "M68010",
    0x02: "M68020",
    0x03: "SPARC",
    0x04: "R3000",
    0x40: "NS32032",
    0x45: "NS32532",
    0x64: "386",
    0x65: "AMD 29K",
    0x66: "386 DYNIX",
    0x67: "ARM",
    0x68: "IBM RT (ROMP AOS)",
    0x83: "Sparclet",
    0x86: "NetBSD/i386",
    0x87: "NetBSD/m68k",
    0x88: "NetBSD/m68k4k",
    0x89: "NetBSD/ns32532",
    0x8A: "NetBSD/SPARC",
    0x8B: "NetBSD/pmax",
    0x8C: "NetBSD/vax 1k",
    0x8D: "NetBSD/alpha",
    0x8E: "NetBSD/mips",
    0x8F: "NetBSD/arm32",
    0x91: "SH3",
    0x94: "PowerPC 64",
    0x95: "NetBSD/powerpc",
    0x96: "NetBSD/vax 4k",
    0x97: "MIPS R2000/R3000",
    0x98: "MIPS R4000/R6000",
    0x99: "OpenBSD/m88k",
    0x9A: "OpenBSD/HPPA",
    0x9B: "SuperH 64-bit",
    0x9C: "NetBSD/sparc64",
    0x9D: "NetBSD/amd64",
    0x9E: "SuperH 32-bit",
    0x9F: "Itanium",
    0xB7: "ARM AARCH64",
    0xB8: "OpenRISC 1000",
    0xB9: "RISC-V",
    0xFF: "Axis ETRAX CRIS",
}


def extract_a_out_info(filetype: str, filename: str) -> object:
    try:
        with open(filename, "rb") as f:
            magic_bytes = f.read(4)
            target = get_target_type(filetype, magic_bytes)
            if target is None:
                return None
            return {"aoutMachineType": target}
    except FileNotFoundError:
        return None


def get_target_type(filetype: str, magic_bytes: bytes) -> Union[str, None]:
    if filetype == "A.OUT big":
        big_endian_magic = (
            int.from_bytes(magic_bytes[:4], byteorder="big", signed=False) >> 16
        ) & 0xFF
        if big_endian_magic in _A_OUT_TARGET_NAME:
            return _A_OUT_TARGET_NAME[big_endian_magic]
    if filetype == "A.OUT little":
        little_endian_magic = (
            int.from_bytes(magic_bytes[:4], byteorder="little", signed=False) >> 16
        ) & 0xFF
        if little_endian_magic in _A_OUT_TARGET_NAME:
            return _A_OUT_TARGET_NAME[little_endian_magic]
    return None



"""setupstepper.py"""

from json import loads
from logging import DEBUG, basicConfig, exception, info
from ntpath import basename
from os import listdir, remove, system
from os.path import exists
from re import sub
from shutil import copy, copy2
from time import sleep
from warnings import simplefilter

from pywinauto.application import Application, WindowSpecification
from pywinauto.findbestmatch import MatchError
from pywinauto.findwindows import find_elements

# Configure logging output
basicConfig(format="%(message)s", encoding="utf-16", level=DEBUG)

# Disable warnings
simplefilter("ignore", category=UserWarning)

# Configurable Strings
DRIVE = "Z:"
UNFILTERED = "results.txt"
FILE_LIST = "files.txt"
FILE_SIGNAL = "done.txt"

# Passed in command line arguments as dictionary (flag: value)
arguments = {"-path": None, "-debug": "off", "-license": "Stepper"}

# List of controls that have elevated priority
desired = ["next", "yes", "finish", "ok", "start", "install", "complete"]
desired.extend(["full"])

# List of controls that threw an error and should not be touched
blacklisted = []


def launch_installer() -> Application:
    """Creates an application object and launches the desired installer based
    on the path supplied in the arguments. If the installer is an msi, it is
    launched with msiexec.

    Args:
        None.

    Returns:
        Application: The application object that started the installer.
    """

    # Create an application object
    app = Application()

    # Start application with proper arguments
    msiprefix = "msiexec /i " if (arguments["-type"] == "msi") else ""
    app = app.start(f'{msiprefix}"{arguments["-path"]}"')

    # Wait for the application to start
    sleep(1.5 if (arguments["-type"] == "msi") else 0.75)

    return app


def get_controls(dialog: WindowSpecification) -> list:
    """Gathers relevant controls from the given dialog and organizes their
    class, name, and ID into a sublist that will be put into the 'controls'
    list.

    Args:
        dialog (WindowSpecification): The dialog to be analyzed.

    Returns:
        list: Contains sublists that contain information about each control.
    """

    controls = []

    # Initiate list of button names that shouldn't be considered
    unwanted_controls = ["Static", "#32770", "CtrlNotifySink", "DirectUIHWND"]
    unwanted_controls.extend(["SysLink", "TNewNotebookPage", "TNewNotebook"])

    edit_number = 1

    for child in dialog.children():
        child_class = child.friendly_class_name()

        # Assign edit control numbers appropriately
        if child_class == "Edit":
            child_class += str(edit_number)
            edit_number += 1

        # Dont include a non-selectable or unwanted control
        if not child.is_enabled() or (child_class in unwanted_controls):
            continue

        # For readability, remove all special characters from non-text boxes
        name = child.texts()[0]

        if not child_class.startswith("Edit"):
            name = sub(r"^\W*|[^0-9a-zA-Z| ]+|\W*$", "", name)

        # For each control, make a sublist with Type, Text, and ID
        controls.append([child_class, name.lower(), child.control_id()])

    return controls


def get_priority_list(ctrl: list) -> list:
    """Creates a parallel list to 'ctrl' whose nth integer value represents
    the priority value of the nth control. Priority values are assigned
    based on various conditions that attempt to generalize the next best
    step that results in successful installation.

    Args:
        ctrl (list): The list of controls returned from get_controls().

    Returns:
        list: Has priority values for each control in its parallel list 'ctrl'.
    """

    num_controls = len(ctrl)
    priority = [0] * num_controls

    for i in range(num_controls):
        classname, text, control_id = ctrl[i]
        is_edit = classname.startswith("Edit")

        # Condition for license RadioButton
        license_cond = "agree" in text or "accept" in text
        license_cond = license_cond and ("not" not in text) and not is_edit

        # Condition for presence of an edit box
        edit_cond = (arguments["-license"] != text) and is_edit
        edit_cond = edit_cond and not text.startswith("c:")
        edit_cond_value = len(text)

        # Condition for control text being in the desired list
        desired_cond = text in desired

        # Condition for controls to avoid
        bad_cond = (control_id in blacklisted) or (text == "cancel")
        bad_cond = bad_cond or ((text == "") and not is_edit)

        # Assign priority value
        if bad_cond:
            priority[i] = 0
        elif edit_cond:
            priority[i] = 4 + edit_cond_value
        elif desired_cond:
            priority[i] = 3
        elif license_cond:
            priority[i] = 2
        else:
            priority[i] = 1

    return priority


def proceed(dialog: WindowSpecification, control: list, p_val: int) -> None:
    """Decides whether to interact with the control or to wait. If the
    installation has started, allow the interaction with any control to exit
    the installer that would disrupt the installation configuration beforehand.
    Buttons are clicked and Edits are scrolled and have their text changed.

    Args:
        dialog (WindowSpecification): The dialog box containing the 'controls'.
        control (list): Attributes of the control with the highest priority.
        p_val (int): The priority value of 'control'.

    Returns:
        None.
    """

    # Decide whether to wait or to advance
    classname, text, control_id = control if (p_val != 0) else ["", "", ""]

    # If a control was the culprit, add it to the blacklist
    on_trial = 0 if "next" in text else control_id
    blacklisted.append(on_trial)

    # Scroll down and change the text in the edit box to the license code
    if classname.startswith("Edit"):
        dialog[classname].scroll(direction="down", amount="end")
        dialog[classname].set_edit_text(arguments["-license"])
        blacklisted.pop()
        blacklisted.append(control_id)

    # Perform button press if we have advanced
    elif classname != "":
        dialog[text].click()
        blacklisted.pop()

        # Add exit option once an install button has been pressed
        if text in ("install", "start"):
            desired.append("exit")

    # Give the UI a chance to load controls if it hasn't
    else:
        blacklisted.pop()
        sleep(0.2)


def step_through(app: Application) -> None:
    """The main loop that sorts the controls from the current dialog and steps
    through the installer. If the '-debug' flag is on, the list of controls are
    dumped in descending order of priority and the user can determine whether
    the next step is done manually or not.

    Args:
        app (Application): The Application object that started the installer.

    Returns:
        None.
    """

    controls = None

    while app.is_process_running():
        try:
            # Get dialog box and retry if none are found
            window = app.top_window().texts()[0]

            if window == "":
                continue

            dialog = app[window]

            # Get controls and sort them. Retry if no controls are found
            controls = get_controls(dialog)
            p_l = get_priority_list(controls)

            if len(p_l) == 0:
                continue

            p_l, controls = (list(t) for t in zip(*sorted(zip(p_l, controls))))
            p_l.reverse()
            controls.reverse()

            # Print controls in debug mode
            auto_step = ""

            if arguments["-debug"] == "on":
                info("Controls for: " + window)
                info("--------------------------------")

                for control in controls:
                    info(f"> {control[0]: <15} {control[1]: <15}")

                auto_step = input("\nENTER to step or type to manually step:")
                info("")

            # Decide what to press next if debug mode allowed it
            if len(auto_step) == 0:
                proceed(dialog, controls[0], p_l[0])

        except (RuntimeError, MatchError, TimeoutError) as err:
            # If a new process has been spawned, latch on to that one instead
            if str(err) == "No windows for that process could be found":
                # Indicates that the installer has unexpectedly terminated
                if controls is not None:
                    break

                info("Ensure that the installer is in focus.")

                if arguments["-debug"] == "on":
                    info(find_elements(active_only=True))

                # Best way to generally latch on to multiprocess installers
                app.connect(active_only=True, found_index=0)
                continue

            exception(f"Error: {err}")

    info("Installer has terminated.")


def handle_transfers() -> None:
    """Read the list of files that the host machine wants and attempt to
    transfer them into the shared folder. 'done.txt' signals to the host
    that all of the files have been acknowledged.

    Args:
        None.

    Returns:
        None.
    """

    # Copy each file into the shared folder
    with open(f"{DRIVE}/{FILE_LIST}", "r", encoding="utf-8") as f_handle:
        for line in f_handle.readlines():
            # The name for each file is changed to a variant of its path
            file = line.strip()
            newname = file[3:].replace("/", "!SEP!")

            try:
                copy2(file, f"{DRIVE}/{newname}")
            except IOError as err:
                exception(err)

    # Signal to the host that each file has finished copying
    with open(FILE_SIGNAL, "w", encoding="utf-8") as f_handle:
        f_handle.write("done")

    copy(FILE_SIGNAL, f"{DRIVE}/{FILE_SIGNAL}")
    remove(FILE_SIGNAL)

    # "Sleep" until the host has finished processing files
    while exists(f"{DRIVE}/{FILE_SIGNAL}"):
        sleep(0.5)


def handle_file(fname: str) -> None:
    """If the file is a txt, it's assumed to be the arguments for the file
    that will be placed into the folder next, which it will run and step
    through. Finally, the file is deleted from the folder. The installer
    is meant to be copied (not moved) into the directory.

    Args:
        fname (str): The name of the file to inspect

    Returns:
        None.
    """

    if arguments["-debug"] == "on":
        info(f"file: {fname}")

    # Handle file transfers
    if basename(fname) == FILE_LIST:
        handle_transfers()
        return

    # Handle installer file
    if fname[-3:] != "txt":
        arguments["-path"] = fname

        try:
            # Open and execute the installer
            app = launch_installer()
            step_through(app)

            # Stop capturing output after a bit
            sleep(10)
            system("taskkill /f /im minifilter.exe")

            # Move results into the shared folder and cleanup
            remove(fname)
            sleep(5)
            copy(UNFILTERED, f"{DRIVE}/{UNFILTERED}")
            remove(UNFILTERED)

            # Wait for results.txt to transfer
            while exists(f"{DRIVE}/{UNFILTERED}"):
                sleep(0.5)

            info(f"{UNFILTERED} file move complete")
        except OSError as err:
            exception(err)
            sleep(0.5)

        return

    # Parse txt into dict as json
    try:
        info("Processing text file as args...")
        with open(fname, "r", encoding="utf-8") as f_handle:
            argstr = f_handle.readlines()[0]
            arguments.update(loads(argstr))

        # Remove file from folder
        remove(fname)
    except IOError as err:
        exception(err)
        sleep(0.5)


def main() -> None:
    """Continously checks to see if any new file has been added to the shared
    folder. If it has, it passes the file path to the handler. This script is
    meant to always run on a windows VM with admin privileges.

    Args:
        None.

    Returns:
        None.
    """

    while True:
        # Retry if the shared folder doesn't exist
        if not exists(f"{DRIVE}/"):
            continue

        try:
            # Check for any files in the shared folder
            files = listdir(f"{DRIVE}/")

            if len(files) == 0:
                continue

            handle_file(f"{DRIVE}/{files[0]}")
        except OSError as err:
            exception(err)


if __name__ == "__main__":
    main()



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from collections.abc import Iterable

from loguru import logger

from surfactant.sbomtypes import SBOM


# TODO for an intermediate SBOM format, have ability to search more efficiently by hashes/filepath/filename
# currently, establishing relationships is something around O(n^2) due to searching entire sbom for matches
def parse_relationships(pluginmanager, sbom: SBOM):
    for i, sw in enumerate(sbom.software):
        logger.info(f"Determining relationship {i + 1}/{len(sbom.software)}")
        # Skip for temporary files/installer that don't have any installPath or metadata to find dependencies with
        if sw.installPath is None or sw.metadata is None:
            continue

        # Find metadata saying what dependencies are used by the software entry
        for md in sw.metadata:
            # handle ELF dependencies, PE imports, and dotNet assembly references using included plugins
            for relationships in pluginmanager.hook.establish_relationships(
                sbom=sbom, software=sw, metadata=md
            ):
                if isinstance(relationships, Iterable):
                    for r in relationships:
                        if not sbom.find_relationship_object(r):
                            sbom.add_relationship(r)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import uuid
from dataclasses import dataclass, field
from typing import List, Optional

from ._file import File
from ._provenance import AnalysisDataProvenance

# pylint: disable=too-many-instance-attributes


@dataclass
class AnalysisData:
    # origin, testName, and testVersion are not optional
    # user must provide the info, not really a "safe" default value
    origin: str
    testName: str
    testVersion: str
    UUID: str = field(default_factory=lambda: str(uuid.uuid4()))
    specificEnvironment: Optional[str] = None
    files: Optional[List[File]] = None
    linksToKnownVulnerabilities: Optional[str] = None
    provenance: Optional[List[AnalysisDataProvenance]] = None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from typing import Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM


@surfactant.plugin.hookimpl
def write_sbom(sbom: SBOM, outfile) -> None:
    # outfile is a file pointer, not a file name
    outfile.write(sbom.to_json(indent=2))


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "cytrics"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from surfactant.plugin.manager import get_plugin_manager
from surfactant.sbomtypes import SBOM, Relationship, Software

sbom = SBOM(
    software=[
        Software(
            UUID="application",
            fileName=["application"],
            installPath=["C:\\application"],
            metadata=[
                {
                    "dotnetAssemblyRef": [{"Name": "samedirlib"}],
                },
                {
                    "dotnetAssemblyRef": [{"Name": "subdirlib"}],
                },
                {
                    "dotnetAssemblyRef": [
                        {
                            "Name": "culturelib",
                            "Culture": "culture",
                        }
                    ],
                },
            ],
        ),
        Software(
            UUID="samedirlib",
            fileName=["samedirlib.dll"],
            installPath=["C:\\samedirlib.dll"],
        ),
        Software(
            UUID="subdirlib",
            fileName=["subdirlib.dll"],
            installPath=["C:\\subdirlib\\subdirlib.dll"],
        ),
        Software(
            UUID="culturelib",
            fileName=["culturelib.dll"],
            installPath=["C:\\culture\\culturelib.dll"],
        ),
    ],
)


def test_same_directory():
    dotnet = get_plugin_manager().get_plugin("surfactant.relationships.dotnet_relationship")
    sw = sbom.software[0]
    md = sw.metadata[0]
    assert dotnet.establish_relationships(sbom, sw, md) == [
        Relationship("application", "samedirlib", "Uses")
    ]


def test_subdir():
    dotnet = get_plugin_manager().get_plugin("surfactant.relationships.dotnet_relationship")
    sw = sbom.software[0]
    md = sw.metadata[1]
    assert dotnet.establish_relationships(sbom, sw, md) == [
        Relationship("application", "subdirlib", "Uses")
    ]


def test_culture():
    dotnet = get_plugin_manager().get_plugin("surfactant.relationships.dotnet_relationship")
    sw = sbom.software[0]
    md = sw.metadata[2]
    assert dotnet.establish_relationships(sbom, sw, md) == [
        Relationship("application", "culturelib", "Uses")
    ]



from pathlib import Path

from surfactant.cmd.cli_commands.cli_base import Cli
from surfactant.plugin.manager import find_io_plugin, get_plugin_manager


class Load(Cli):
    """
    A class that implements the surfactant cli load functionality

    Attributes:
        input_format (str): The format for input sboms
    """

    def __init__(self, *args, input_format, **kwargs):
        """Executes the load class constructor

        Args:
            input_format (str): The format of the sbom being loaded
        """
        self.input_format = input_format
        super().__init__(*args, **kwargs)

    def execute(self, input_file):
        """Executes the main functionality of the load class

        Args:
            input_file: The sbom load into the cli
        """
        pm = get_plugin_manager()
        input_reader = find_io_plugin(pm, self.input_format, "read_sbom")
        self.sbom = input_reader.read_sbom(input_file)

        serialized_sbom = Cli.serialize(self.sbom)
        with open(Path(self.data_dir, self.sbom_filename), "wb") as f:
            f.write(serialized_sbom)



from typing import Any, Dict

import javatools.jarinfo

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software

# TODO: Add documentation about how to install javatools
# swig and libssl-dev needs to be installed on Ubuntu
# https://gitlab.com/m2crypto/m2crypto/-/blob/master/INSTALL.rst


def supports_file(filetype: str) -> bool:
    return filetype in ("JAVACLASS", "JAR", "WAR", "EAR")


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    return extract_java_info(filename, filetype)


# Map from internal major number to Java SE version
# https://docs.oracle.com/javase/specs/jvms/se20/html/jvms-4.html#jvms-4.1-200-B.2
_JAVA_VERSION_MAPPING = {
    45: "1.1",
    46: "1.2",
    47: "1.3",
    48: "1.4",
    49: "5.0",
    50: "6",
    51: "7",
    52: "8",
    53: "9",
    54: "10",
    55: "11",
    56: "12",
    57: "13",
    58: "14",
    59: "15",
    60: "16",
    61: "17",
    62: "18",
    63: "19",
    64: "20",
}


def handle_java_class(info: Dict[str, Any], class_info: javatools.JavaClassInfo):
    # This shouldn't happen but just in-case it does don't overwrite information
    if class_info.get_this() in info["javaClasses"]:
        return
    info["javaClasses"][class_info.get_this()] = {}
    add_to = info["javaClasses"][class_info.get_this()]
    (major_version, _) = class_info.get_version()
    if major_version in _JAVA_VERSION_MAPPING:
        add_to["javaMinSEVersion"] = _JAVA_VERSION_MAPPING[major_version]
    add_to["javaExports"] = [*class_info.get_provides()]
    # I've seen this fail for some reason; catch errors on it and just ignore
    # them if it fails
    try:
        add_to["javaImports"] = [*class_info.get_requires()]
    except IndexError:
        # Should this be set to "Unknown" or similar?
        add_to["javaImports"] = []


def extract_java_info(filename: str, filetype: str) -> object:
    info: Dict[str, Any] = {"javaClasses": {}}
    if filetype in ("JAR", "EAR", "WAR"):
        with javatools.jarinfo.JarInfo(filename) as jarinfo:
            for class_ in jarinfo.get_classes():
                handle_java_class(info, jarinfo.get_classinfo(class_))
    elif filetype == "JAVACLASS":
        with open(filename, "rb") as f:
            class_info = javatools.JavaClassInfo()
            class_info.unpack(javatools.unpack(f))
        handle_java_class(info, class_info)
    return info



import json
import re

import requests

from surfactant.configmanager import ConfigManager


def get_test_file():
    url = "https://cdnjs.cloudflare.com/ajax/libs/select2/3.5.4/select2.min.js"
    response = requests.get(url)
    if response.status_code == 200:
        with open("testFile.js", "w") as js:
            js.write(response.text)


def find_js_match(expressions: dict, filename: str) -> str:
    for name, library in expressions.items():
        if "filename" in library:
            for pattern in library["filename"]:
                if re.search(pattern, filename):
                    return name
    try:
        with open(filename, "r") as jsfile:
            contents = jsfile.read()
        for name, library in expressions.items():
            if "filecontent" in library:
                for pattern in library["filecontent"]:
                    if re.search(pattern, contents):
                        return name
    except FileNotFoundError:
        print(f"File not found: {filename}")
    return None


get_test_file()
json_file_path = ConfigManager().get_data_dir_path() / "infoextractors" / "js_library_patterns.json"
with open(json_file_path, "r") as f:
    patterns = json.load(f)

library_name = find_js_match(patterns, "testFile.js")
print(library_name)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from .cli_base import Cli
from .cli_load import Load
from .cli_save import Save

__all__ = ["Load", "Save", "Cli"]



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import pathlib

import pytest

from surfactant.cmd.cli import cli_add, cli_find
from surfactant.cmd.cli_commands import Cli
from surfactant.sbomtypes import SBOM, Relationship


@pytest.fixture(name="test_sbom")
def fixture_test_sbom():
    with open(pathlib.Path(__file__).parent / "../data/sample_sboms/helics_sbom.json", "r") as f:
        sbom = SBOM.from_json(f.read())
        return sbom


def _compare_sboms(one: SBOM, two: SBOM) -> bool:
    # Sort software list
    one.software = sorted(one.software, key=lambda x: x.UUID)
    two.software = sorted(two.software, key=lambda x: x.UUID)

    # Sort hardware list
    one.hardware = sorted(one.hardware, key=lambda x: x.UUID)
    two.hardware = sorted(two.hardware, key=lambda x: x.UUID)

    # Sort system list
    one.systems = sorted(one.systems, key=lambda x: x.UUID)
    two.systems = sorted(two.systems, key=lambda x: x.UUID)

    # Sort relationship list
    one.relationships = sorted(one.relationships, key=lambda x: x.yUUID)
    two.relationships = sorted(two.relationships, key=lambda x: x.yUUID)

    return one.to_dict() == two.to_dict()


bad_sbom = SBOM(
    {
        "software": [
            {
                "UUID": "477da45b-bb38-450e-93f7-e525aaaa6862",
                "name": None,
                "size": 16367492,
                "fileName": ["helics.tar.gz"],
                "installPath": [],
                "containerPath": [],
                "captureTime": 1689186121,
                "version": "",
                "vendor": [],
                "description": "",
                "sha1": "0d21026ee953eeaa31cafef5118be56f46867267",
                "sha256": "f41ca6f7c447225df3a7eef754d303d22cf877586735fb2d56d1eb15bf1daed9",
                "md5": "5fbf80df5004db2f0ce1f78b524024fe",
                "relationshipAssertion": "Unknown",
                "comments": "",
                "supplementaryFiles": [],
                "provenance": None,
                "recordedInstitution": "LLNL",
                "components": [],
                "bad_key": 1.24553,
            }
        ]
    }
)


def test_find_by_sha256(test_sbom):
    out_bom = cli_find().execute(
        test_sbom, sha256="f41ca6f7c447225df3a7eef754d303d22cf877586735fb2d56d1eb15bf1daed9"
    )
    assert len(out_bom.software) == 1
    assert (
        out_bom.software[0].sha256
        == "f41ca6f7c447225df3a7eef754d303d22cf877586735fb2d56d1eb15bf1daed9"
    )


def test_find_by_multiple_hashes(test_sbom):
    out_bom = cli_find().execute(
        test_sbom,
        sha256="f41ca6f7c447225df3a7eef754d303d22cf877586735fb2d56d1eb15bf1daed9",
        md5="5fbf80df5004db2f0ce1f78b524024fe",
    )
    assert len(out_bom.software) == 1
    assert (
        out_bom.software[0].sha256
        == "f41ca6f7c447225df3a7eef754d303d22cf877586735fb2d56d1eb15bf1daed9"
    )


def test_find_by_mismatched_hashes(test_sbom):
    out_bom = cli_find().execute(
        test_sbom,
        sha256="f41ca6f7c447225df3a7eef754d303d22cf877586735fb2d56d1eb15bf1daed9",
        md5="2ff380e740d2eb09e5d67f6f2cd17636",
    )
    assert len(out_bom.software) == 0


def test_find_by_containerPath(test_sbom):
    out_bom = cli_find().execute(test_sbom, containerpath="477da45b-bb38-450e-93f7-e525aaaa6862/")
    assert len(out_bom.software) == 7


def test_find_with_malformed_sbom():
    out_bom = cli_find().execute(bad_sbom, bad_key=1.24553)  # Unsupported Type
    assert len(out_bom.software) == 0
    out_bom = cli_find().execute(bad_sbom, bad_key="testing")  # Supported Type
    assert len(out_bom.software) == 0


def test_find_with_bad_filter():
    out_bom = cli_find().execute(bad_sbom, bad_filter="testing")  # Supported Type
    assert len(out_bom.software) == 0
    out_bom = cli_find().execute(bad_sbom, bad_filter=1.234)  # Unsupported Type
    assert len(out_bom.software) == 0


def test_add_by_file(test_sbom):
    previous_software_len = len(test_sbom.software)
    out_bom = cli_add().execute(
        test_sbom, file=pathlib.Path(__file__).parent / "../data/a_out_files/big_m68020.aout"
    )
    assert len(out_bom.software) == previous_software_len + 1
    assert (
        out_bom.software[8].sha256
        == "9e125f97e5f180717096c57fa2fdf06e71cea3e48bc33392318643306b113da4"
    )


def test_add_entry(test_sbom):
    entry = {
        "UUID": "6b50c545-3e07-4aec-bbb0-bae07704143a",
        "name": "Test Aout File",
        "size": 4,
        "fileName": ["big_m68020.aout"],
        "installPath": [],
        "containerPath": [],
        "captureTime": 1715726918,
        "sha1": "fbf8688fbe1976b6f324b0028c4b97137ae9139d",
        "sha256": "9e125f97e5f180717096c57fa2fdf06e71cea3e48bc33392318643306b113da4",
        "md5": "e8d3808a4e311a4262563f3cb3a31c3e",
        "comments": "This is a test entry.",
    }
    previous_software_len = len(test_sbom.software)
    out_bom = cli_add().execute(test_sbom, entry=entry)
    assert len(out_bom.software) == previous_software_len + 1
    assert (
        out_bom.software[8].sha256
        == "9e125f97e5f180717096c57fa2fdf06e71cea3e48bc33392318643306b113da4"
    )


def test_add_relationship(test_sbom):
    relationship = {
        "xUUID": "455341bb-2739-4918-9805-e1a93e27e2a4",
        "yUUID": "e286a415-6c6b-427d-9fe6-d7dbb0486f7d",
        "relationship": "Uses",
    }
    previous_rel_len = len(test_sbom.relationships)
    out_bom = cli_add().execute(test_sbom, relationship=relationship)
    assert len(out_bom.relationships) == previous_rel_len + 1
    test_sbom.relationships.discard(Relationship(**relationship))


def test_add_installpath(test_sbom):
    containerPathPrefix = "477da45b-bb38-450e-93f7-e525aaaa6862/"
    installPathPrefix = "/bin/"
    out_bom = cli_add().execute(test_sbom, installpath=(containerPathPrefix, installPathPrefix))
    for sw in out_bom.software:
        if containerPathPrefix in sw.containerPath:
            assert installPathPrefix in sw.installPath


def test_cli_base_serialization(test_sbom):
    serialized = Cli.serialize(test_sbom)
    deserialized = Cli.deserialize(serialized)
    assert test_sbom == deserialized
    assert _compare_sboms(test_sbom, deserialized)



import hashlib
import sys
from pathlib import Path

import click
from loguru import logger

from surfactant.cmd.cli_commands import Load, Save
from surfactant.configmanager import ConfigManager
from surfactant.plugin.manager import find_io_plugin, get_plugin_manager
from surfactant.sbomtypes._relationship import Relationship
from surfactant.sbomtypes._sbom import SBOM
from surfactant.sbomtypes._software import Software


@click.argument("sbom", type=click.File("r"), required=True)
@click.option(
    "--input_format",
    is_flag=False,
    default=ConfigManager().get(
        "core", "input_format", fallback="surfactant.input_readers.cytrics_reader"
    ),
    help="SBOM input format, assumes that all input SBOMs being merged have the same format, options=[cytrics|cyclonedx|spdx]",
)
@click.command("load")
def handle_cli_load(sbom, input_format):
    "CLI command to load supplied SBOM into cli"
    Load(input_format=input_format).execute(sbom)


@click.argument("sbom", type=click.File("r"), required=True)
@click.option("--file", is_flag=False, help="File of the entry to find")
@click.option("--sha256", is_flag=False, type=str, help="sha256 hash of the entry to find")
@click.option("--uuid", is_flag=False, type=str, help="uuid of the entry to find")
@click.option(
    "--installPath",
    is_flag=False,
    type=str,
    help="Matches all entries with an install path or partial install path match",
)
@click.option(
    "--containerPath",
    is_flag=False,
    type=str,
    help="Matches all entries with a container path or partial container path match",
)
@click.option(
    "--output_format",
    is_flag=False,
    default="surfactant.output.cytrics_writer",
    help="SBOM output format, options=[cytrics|csv|spdx|cyclonedx]",
)
@click.option(
    "--input_format",
    is_flag=False,
    default="surfactant.input_readers.cytrics_reader",
    help="SBOM input format, assumes that all input SBOMs being merged have the same format, options=[cytrics|cyclonedx|spdx]",
)
@click.command("find")
def handle_cli_find(sbom, output_format, input_format, **kwargs):
    "CLI command to find specific entry(s) within a supplied SBOM"
    pm = get_plugin_manager()
    output_writer = find_io_plugin(pm, output_format, "write_sbom")
    input_reader = find_io_plugin(pm, input_format, "read_sbom")
    in_sbom = input_reader.read_sbom(sbom)

    # Remove None values
    filtered_kwargs = dict({(k, v) for k, v in kwargs.items() if v is not None})
    out_sbom = cli_find().execute(in_sbom, **filtered_kwargs)
    if not out_sbom.software:
        logger.warning("No software matches found with given parameters.")
    output_writer.write_sbom(out_sbom, sys.stdout)


@click.argument("sbom", required=True)
@click.option(
    "--output",
    default=None,
    is_flag=False,
    help="Specifies the file to output new sbom. Default replaces the input file.",
)
@click.option("--file", is_flag=False, help="Adds entry for file to sbom")
@click.option("--relationship", is_flag=False, type=str, help="Adds relationship to sbom")
@click.option("--entry", is_flag=False, type=str, help="Adds software entry to sbom")
@click.option(
    "--installPath",
    is_flag=False,
    type=str,
    nargs=2,
    help="Adds new installPath by finding and replacing a containerPath prefix (1st arg) with a new prefix (2nd arg)",
)
@click.option(
    "--output_format",
    is_flag=False,
    default="surfactant.output.cytrics_writer",
    help="SBOM output format, options=[cytrics|csv|spdx|cyclonedx]",
)
@click.option(
    "--input_format",
    is_flag=False,
    default="surfactant.input_readers.cytrics_reader",
    help="SBOM input format, options=[cytrics|cyclonedx|spdx]",
)
@click.command("add")
def handle_cli_add(sbom, output, output_format, input_format, **kwargs):
    "CLI command to add specific entry(s) to a supplied SBOM"
    pm = get_plugin_manager()
    output_writer = find_io_plugin(pm, output_format, "write_sbom")
    input_reader = find_io_plugin(pm, input_format, "read_sbom")
    with open(Path(sbom), "r") as f:
        in_sbom = input_reader.read_sbom(f)
    # Remove None values
    filtered_kwargs = dict({(k, v) for k, v in kwargs.items() if v is not None})
    out_sbom = cli_add().execute(in_sbom, **filtered_kwargs)
    # Write to the input file if no output specified
    if output is None:
        with open(Path(sbom), "w") as f:
            output_writer.write_sbom(out_sbom, f)
    else:
        try:
            with open(Path(output), "w") as f:
                output_writer.write_sbom(out_sbom, f)
        except OSError as e:
            logger.error(f"Could not open file {output} in write mode - {e}")


@click.argument("sbom", type=click.File("r"), required=True)
@click.command("edit")
def handle_cli_edit(sbom, output_format, input_format, **kwargs):
    "CLI command to edit specific entry(s) in a supplied SBOM"


@click.argument("outfile", type=click.File("w"), required=True)
@click.option(
    "--output_format",
    is_flag=False,
    default=ConfigManager().get(
        "core", "output_format", fallback="surfactant.output.cytrics_writer"
    ),
    help="SBOM output format, options=[cytrics|csv|spdx|cyclonedx]",
)
@click.command("save")
def handle_cli_save(outfile, output_format):
    "CLI command to save SBOM to a user specified file"
    Save(output_format=output_format).execute(outfile)


class cli_add:
    """
    A class that implements the surfactant cli add functionality

    Attributes:
    match_functions         A dictionary of functions that provide matching functionality for given SBOM fields (i.e. uuid, sha256, installpath, etc)
    camel_case_conversions  A dictionary of string conversions from all lowercase to camelcase. Used to convert python click options to match the SBOM attribute's case
    sbom                    An internal record of sbom entries the class adds to as it finds more matches.
    """

    camel_case_conversions: dict
    match_functions: dict
    sbom: SBOM

    def __init__(self):
        """Initializes the cli_add class"""
        self.match_functions = {
            "relationship": self.add_relationship,
            "file": self.add_file,
            "installPath": self.add_installpath,
            "entry": self.add_entry,
        }
        self.camel_case_conversions = {
            "uuid": "UUID",
            "filename": "fileName",
            "installpath": "installPath",
            "capturetime": "captureTime",
            "relationshipassertion": "relationshipAssertion",
        }

    def handle_kwargs(self, kwargs: dict) -> dict:
        converted_kwargs = {}
        for k, v in kwargs.items():  # Convert key values to camelcase where appropriate
            key = self.camel_case_conversions[k] if k in self.camel_case_conversions else k
            converted_kwargs[key] = v
        return converted_kwargs

    def execute(self, input_sbom: SBOM, **kwargs):
        """Executes the main functionality of the cli_find class
        param: input_sbom   The sbom to add entries to
        param: kwargs:      Dictionary of key/value pairs indicating what features to match on
        """
        converted_kwargs = self.handle_kwargs(kwargs)
        self.sbom = input_sbom

        for key, value in converted_kwargs.items():
            if key in self.match_functions:
                self.match_functions[key](value)
            else:
                logger.warning(f"Paramter {key} is not supported")
        return self.sbom

    def add_relationship(self, value: dict) -> bool:
        self.sbom.add_relationship(Relationship(**value))

    def add_file(self, path):
        self.sbom.software.append(Software.create_software_from_file(path))

    def add_entry(self, entry):
        self.sbom.software.append(Software.from_dict(entry))

    def add_installpath(self, prefixes: tuple):
        cleaned_prefixes = (p.rstrip("/") for p in prefixes)
        containerPathPrefix, installPathPrefix = cleaned_prefixes
        for sw in self.sbom.software:
            for path in sw.containerPath:
                if containerPathPrefix in path:
                    sw.installPath.append(path.replace(containerPathPrefix, installPathPrefix))


class cli_find:
    """
    A class that implements the surfactant cli find functionality

    Attributes:
    match_functions         A dictionary of functions that provide matching functionality for given SBOM fields (i.e. uuid, sha256, installpath, etc)
    camel_case_conversions  A dictionary of string conversions from all lowercase to camelcase. Used to convert python click options to match the SBOM attribute's case
    sbom                    An internal record of sbom entries the class adds to as it finds more matches.
    """

    match_functions: dict
    camel_case_conversions: dict
    sbom: SBOM

    def __init__(self):
        """Initializes the cli_find class"""
        self.match_functions = {
            int: self.match_single_value,
            str: self.match_single_value,
            list: self.match_array_value,
            dict: self.match_dict_value,
            float: self.match_none_or_unhandled,
            tuple: self.match_none_or_unhandled,
            type(None): self.match_none_or_unhandled,
        }
        self.camel_case_conversions = {
            "uuid": "UUID",
            "filename": "fileName",
            "containerpath": "containerPath",
            "installpath": "installPath",
            "capturetime": "captureTime",
            "relationshipassertion": "relationshipAssertion",
        }
        self.sbom = SBOM()

    def handle_kwargs(self, kwargs: dict) -> dict:
        converted_kwargs = {}
        for k, v in kwargs.items():  # Convert key values to camelcase where appropriate
            if k == "file":
                sha256, sha1, md5 = self._calculate_hashes(v, sha256=True, sha1=True, md5=True)
                v = {"sha256": sha256, "sha1": sha1, "md5": md5}
            key = self.camel_case_conversions[k] if k in self.camel_case_conversions else k
            converted_kwargs[key] = v
        return converted_kwargs

    def execute(self, input_sbom: SBOM, **kwargs):
        """Executes the main functionality of the cli_find class
        param: input_sbom   The sbom to find matches within
        param: kwargs:      Dictionary of key/value pairs indicating what features to match on
        """
        converted_kwargs = self.handle_kwargs(kwargs)

        for sw in input_sbom.software:
            match = True
            for k, v in converted_kwargs.items():
                if k == "file":
                    entry_value = {"sha256": sw.sha256, "sha1": sw.sha1, "md5": sw.md5}
                else:
                    entry_value = vars(sw)[k] if k in vars(sw) else None
                if not self.match_functions[type(entry_value)](entry_value, v):
                    match = False
                    break
            if match:
                self.sbom.add_software(sw)
        return self.sbom

    def match_single_value(self, first, second) -> bool:
        """Matches sbom entry on single value
        param: first   The entry value to match
        param: second: The value to match first to
        returns:       bool, True if a match, False if not
        """
        if first == second:
            return True
        return False

    def match_array_value(self, array, value) -> bool:
        """Matches sbom entry on array value. Will match if value is contained in any of the array values.
        param: entry    The entry array to match
        param: value:   The value to find in array
        returns:        bool, True if a match, False if not
        """
        if any(value in entry for entry in array):
            return True
        return False

    def match_dict_value(self, d1: dict, d2: dict) -> bool:
        """Matches dictonary values. Will match if two dictionaries have any k,v pairs in common. Used for file hash comparison.
        param: d1       The first dictionary of values
        param: d2:      The 2nd dictionary of values to find
        returns:        bool, True if a match, False if not
        """
        if set(d1.items()).intersection(set(d2.items())):
            return True
        return False

    def match_none_or_unhandled(self, value, match):
        """Default match function if no key value found in SBOM or match type unknown/unhandled
        param: value    Should only be None
        param: match:   Value that would have been matched
        returns:        False
        """
        logger.debug(f"SBOM entry_value of type={type(value)} is not currently handled.")
        return False

    def _calculate_hashes(self, file, sha256=False, sha1=False, md5=False):
        """Helper function to calculate hashes on a given file.
        param: file      The file to calculate hashes on
        param: sha256:   Bool to decide if sha256 hash should be calculated
        param: sha1:     Bool to decide if sha1 hash should be calculated
        param: md5:      Bool to decide if md5 hash should be calculated
        returns:         str, str, str, Hashes calculated, None for those that aren't calculated
        """
        sha256_hash, sha1_hash, md5_hash = None, None, None
        with open(file, "rb") as f:
            if sha256:
                sha256_hash = hashlib.sha256(f.read()).hexdigest()
                f.seek(0)
            if sha1:
                sha1_hash = hashlib.sha1(f.read()).hexdigest()
                f.seek(0)
            if md5:
                md5_hash = hashlib.md5(f.read()).hexdigest()
        return sha256_hash, sha1_hash, md5_hash



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import os
import stat
import sys
from hashlib import md5, sha1, sha256


def get_file_info(filename):
    """Get information about a file.

    Args:
        filename (str): Name of file.

    Returns:
        Optional[dict]: Dictionary that contains info about the file.
    """
    try:
        fstats = os.stat(filename)
    except FileNotFoundError:
        return None

    filehidden = False
    # stat.UF_HIDDEN (file shouldn't be shown in GUI macOS 10.5+)
    if hasattr(fstats, "st_flags"):
        filehidden = bool(fstats.st_flags & stat.UF_HIDDEN)
    # stat.FILE_ATTRIBUTE_HIDDEN (shouldn't be shown on Windows)
    if hasattr(fstats, "st_file_attributes"):
        filehidden = bool(fstats.st_file_attributes & stat.FILE_ATTRIBUTE_HIDDEN)
    # Should consider if symlinks on Linux/macOS/Windows, or reparse points on Windows need to be recognized and handled
    # lstat() does not follow symbolic links
    # fstats.st_reparse_tag from os.lstat()
    # stat.IO_REPARSE_TAG_SYMLINK
    # stat.IO_REPARSE_TAG_MOUNT_POINT
    return {
        "size": fstats.st_size,
        "accesstime": fstats.st_atime,
        "modifytime": fstats.st_mtime,
        "createtime": fstats.st_ctime,
        "filemode": stat.filemode(fstats.st_mode),
        "filehidden": filehidden,
    }


def calc_file_hashes(filename):
    """Calculate hashes for a file specified.

    Args:
        filename (str): Name of file.

    Returns:
        Optional[dict]: Dictionary with the sha256, sha1, and md5 hashes of the file.
    """
    sha256_hash = sha256()
    sha1_hash = sha1()
    # hashlib.md5 usedforsecurity flag was added in Python 3.9
    if sys.version_info >= (3, 9):
        # avoid error with FIPS-compliant OpenSSL library builds complaining about md5
        md5_hash = md5(usedforsecurity=False)
    else:
        md5_hash = md5()
    b = bytearray(4096)
    mv = memoryview(b)
    try:
        with open(filename, "rb", buffering=0) as f:
            while n := f.readinto(mv):
                sha256_hash.update(mv[:n])
                sha1_hash.update(mv[:n])
                md5_hash.update(mv[:n])
    except FileNotFoundError:
        return None
    return {
        "sha256": sha256_hash.hexdigest(),
        "sha1": sha1_hash.hexdigest(),
        "md5": md5_hash.hexdigest(),
    }


def sha256sum(filename):
    """Calculate sha256 hash for the file specified. May throw a FileNotFound exception.

    Args:
        filename (str): Name of file.

    Returns:
        Optional[str]: The sha256 hash of the file.

    Raises:
        FileNotFoundError: If the given filename could not be found.
    """
    h = sha256()
    with open(filename, "rb") as f:
        # Reading is buffered by default (https://docs.python.org/3/library/functions.html#open)
        chunk = f.read(h.block_size)
        while chunk:
            h.update(chunk)
            chunk = f.read(h.block_size)
    return h.hexdigest()



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import pathlib
from collections.abc import Iterable
from typing import List, Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Relationship, Software

from ._internal.windows_utils import find_installed_software


def has_required_fields(metadata) -> bool:
    # dotnetAssemblyRef must present, otherwise we have no info on .NET imports
    return "dotnetAssemblyRef" in metadata


@surfactant.plugin.hookimpl
def establish_relationships(
    sbom: SBOM, software: Software, metadata
) -> Optional[List[Relationship]]:
    if not has_required_fields(metadata):
        return None

    relationships: List[Relationship] = []
    dependent_uuid = software.UUID
    dnName = None
    dnCulture = None
    dnVersion = None
    if "dotnetAssembly" in metadata:
        dnAssembly = metadata["dotnetAssembly"]
        if "Name" in dnAssembly:
            dnName = dnAssembly["Name"]
        if "Culture" in dnAssembly:
            dnCulture = dnAssembly["Culture"]
        if "Version" in dnAssembly:
            dnVersion = dnAssembly["Version"]

    # get additional probing paths if they exist
    dnProbingPaths = None
    dnDependentAssemblies = None

    windowsAppConfig = None
    windowsManifest = None
    if "appConfigFile" in metadata:
        windowsAppConfig = metadata["appConfigFile"]
    if "manifestFile" in metadata:
        windowsManifest = metadata["manifestFile"]

    if windowsAppConfig:
        if "runtime" in windowsAppConfig:
            wac_runtime = windowsAppConfig["runtime"]
            if "assemblyBinding" in wac_runtime:
                wac_asmbinding = wac_runtime["assemblyBinding"]
                if "dependentAssembly" in wac_asmbinding:
                    dnDependentAssemblies = wac_asmbinding["dependentAssembly"]
                if "probing" in wac_asmbinding:
                    wac_probing = wac_asmbinding["probing"]
                    if "privatePath" in wac_probing:
                        wac_paths = wac_probing["privatePath"]
                        for path in wac_paths.split(";"):
                            if dnProbingPaths is None:
                                dnProbingPaths = []
                            dnProbingPaths.append(pathlib.PureWindowsPath(path).as_posix())

    # https://learn.microsoft.com/en-us/dotnet/core/dependency-loading/loading-unmanaged
    # 1. Check the active AssemblyLoadContext cache
    # 2. Calling the import resolver set by the setDllImportResolver function
    #    - a. Example using SetDllImportResolver: https://learn.microsoft.com/en-us/dotnet/standard/native-interop/native-library-loading
    #    - b. Checks PInvoke's or Assembly's DefaultDllImportSearchPathsAttribute, then the assembly's directory, then LoadLibraryEx with LOAD_WITH_ALTERED_SEARCH_PATH flag
    #       - This attribute has no effect on non-Windows platforms / Mono runtime
    #       - https://learn.microsoft.com/en-us/dotnet/api/system.runtime.interopservices.defaultdllimportsearchpathsattribute?view=net-7.0
    #       - i. This has a "Paths" property which is a bitwise combination of paths specified in ii:
    #       - ii. https://learn.microsoft.com/en-us/dotnet/api/system.runtime.interopservices.dllimportsearchpath?view=net-7.0
    # 3. The active AssemblyLoadContext calls its LoadUnmanagedDll function (Default behavior is the same as AssemblyRef probing?)
    #    - a. Can be overridden, but the default implementation returns IntPtr.Zero, which tells the runtime to load with its default policy.
    #    - b. https://learn.microsoft.com/en-us/dotnet/api/system.runtime.loader.assemblyloadcontext.loadunmanageddll?view=net-7.0
    # 4. Run default unmanaged library probing logic by parsing *.deps.json probing properties
    #    - a. If the json file isn't present, assume the calling assembly's directory contains the library
    #    - b. https://learn.microsoft.com/en-us/dotnet/core/dependency-loading/default-probing#unmanaged-native-library-probing
    if "dotnetImplMap" in metadata:
        for asmRef in metadata["dotnetImplMap"]:
            if "Name" not in asmRef:
                continue
            refName = asmRef["Name"]

            # Check absolute path against entries in software
            if is_absolute_path(refName):
                ref_abspath = pathlib.PureWindowsPath(refName)
                for e in sbom.software:
                    if e.installPath is None:
                        continue
                    if isinstance(e.installPath, Iterable):
                        for ifile in e.installPath:
                            if ref_abspath == pathlib.PureWindowsPath(ifile):
                                relationships.append(Relationship(dependent_uuid, e.UUID, "Uses"))
                continue

            probedirs = []
            if isinstance(software.installPath, Iterable):
                for ipath in software.installPath:
                    probedirs.append(pathlib.PureWindowsPath(ipath).parent.as_posix())
            # Construct a list of combinations specified in (2.a)
            # Refer to Issue #80 - Need to verify that this conforms with cross-platform behavior
            combinations = [refName]
            if not (refName.endswith(".dll") or refName.endswith(".exe")):
                combinations.append(f"{refName}.dll")
            combinations.extend(
                [
                    f"{refName}.so",
                    f"{refName}.dylib",
                    f"lib{refName}.so",
                    f"lib{refName}.dylib",
                    f"lib{refName}",
                ]
            )
            # On Linux, if the libname ends with .so or has .so. then version variations are tried
            # Refer to Issue #79 - Need regex matching for version variations
            for e in find_installed_software(sbom, probedirs, combinations):
                dependency_uuid = e.UUID
                relationships.append(Relationship(dependent_uuid, dependency_uuid, "Uses"))

    # https://learn.microsoft.com/en-us/dotnet/framework/deployment/how-the-runtime-locates-assemblies
    # 1. Determine correct assembly version using configuration files (binding redirects, code location, etc)
    # 2. Check if assembly name bound before; if it is use previously loaded assembly
    # 3. Check global assembly cache (%windir%\Microsoft.NET\assembly in .NET framework 4, %windir%\assembly previously)
    # 4. Probe for assembly:
    # - a. Check for <codeBase> element in app config; check the given location and if assembly found great no probing; otherwise fail without probing
    # - b. If there is no <codeBase> element, begin probing using
    #    - application base + culture + assembly name directories
    #    - privatePath directories from a probing element, combined with culture/appbase/assemblyname (done before the standard probing directories)
    #    - the location of the calling assembly may be used as a hint for where to find the referenced assembly
    if "dotnetAssemblyRef" in metadata:
        for asmRef in metadata["dotnetAssemblyRef"]:
            refName = None
            refVersion = None
            refCulture = None
            if "Name" in asmRef:
                refName = asmRef["Name"]
            else:
                continue  # no name means we have no assembly to search for
            if "Culture" in asmRef:
                refCulture = asmRef["Culture"]
            if "Version" in asmRef:
                refVersion = asmRef["Version"]

            # check if codeBase element exists for this assembly in appconfig
            if dnDependentAssemblies is not None:
                for depAsm in dnDependentAssemblies:
                    # dependent assembly object contains info on assembly id and binding redirects that with a better internal SBOM
                    # representation could be used to also verify the right assembly is being found
                    if "codeBase" in depAsm:
                        if "href" in depAsm["codeBase"]:
                            codebase_href = depAsm["codeBase"]["href"]
                            # strong named assembly can be anywhere on intranet or Internet
                            if (
                                codebase_href.startswith("http://")
                                or codebase_href.startswith("https://")
                                or codebase_href.startswith("file://")
                            ):
                                # codebase references a url; interesting for manual analysis/gathering additional files, but not supported by surfactant yet
                                pass
                            else:
                                # most likely a private assembly, so path must be relative to application's directory
                                if isinstance(software.installPath, Iterable):
                                    for install_filepath in software.installPath:
                                        install_basepath = pathlib.PureWindowsPath(
                                            install_filepath
                                        ).parent.as_posix()
                                        cb_filepath = pathlib.PureWindowsPath(
                                            install_basepath, codebase_href
                                        )
                                        cb_file = cb_filepath.name
                                        cb_path = [cb_filepath.parent.as_posix()]
                                        for e in find_installed_software(sbom, cb_path, cb_file):
                                            dependency_uuid = e.UUID
                                            relationships.append(
                                                Relationship(
                                                    dependent_uuid,
                                                    dependency_uuid,
                                                    "Uses",
                                                )
                                            )

            # continue on to probing even if codebase element was found, since we can't guarantee the assembly identity required by the codebase element
            # get the list of paths to probe based on locations software is installed, assembly culture, assembly name, and probing paths from appconfig file
            probedirs = get_dotnet_probedirs(software, refCulture, refName, dnProbingPaths)
            for e in find_installed_software(sbom, probedirs, refName + ".dll"):
                dependency_uuid = e.UUID
                relationships.append(Relationship(dependent_uuid, dependency_uuid, "Uses"))
                # logging assemblies not found would be nice but is a lot of noise as it mostly just prints system/core .NET libraries
    return relationships


def is_absolute_path(fname: str) -> bool:
    givenpath = pathlib.PureWindowsPath(fname)
    return givenpath.is_absolute()


# construct a list of directories to probe for establishing dotnet relationships
def get_dotnet_probedirs(software: Software, refCulture, refName, dnProbingPaths):
    probedirs = []
    # probe for the referenced assemblies
    if isinstance(software.installPath, Iterable):
        for install_filepath in software.installPath:
            install_basepath = pathlib.PureWindowsPath(install_filepath).parent.as_posix()
            if refCulture is None or refCulture == "":
                # [application base] / [assembly name].dll
                # [application base] / [assembly name] / [assembly name].dll
                probedirs.append(pathlib.PureWindowsPath(install_basepath).as_posix())
                probedirs.append(pathlib.PureWindowsPath(install_basepath, refName).as_posix())
                if dnProbingPaths is not None:
                    # add probing private paths
                    for path in dnProbingPaths:
                        # [application base] / [binpath] / [assembly name].dll
                        # [application base] / [binpath] / [assembly name] / [assembly name].dll
                        probedirs.append(pathlib.PureWindowsPath(install_basepath, path).as_posix())
                        probedirs.append(
                            pathlib.PureWindowsPath(install_basepath, path, refName).as_posix()
                        )
            else:
                # [application base] / [culture] / [assembly name].dll
                # [application base] / [culture] / [assembly name] / [assembly name].dll
                probedirs.append(pathlib.PureWindowsPath(install_basepath, refCulture).as_posix())
                probedirs.append(
                    pathlib.PureWindowsPath(install_basepath, refName, refCulture).as_posix()
                )
                if dnProbingPaths is not None:
                    # add probing private paths
                    for path in dnProbingPaths:
                        # [application base] / [binpath] / [culture] / [assembly name].dll
                        # [application base] / [binpath] / [culture] / [assembly name] / [assembly name].dll
                        probedirs.append(
                            pathlib.PureWindowsPath(install_basepath, path, refCulture).as_posix()
                        )
                        probedirs.append(
                            pathlib.PureWindowsPath(
                                install_basepath, path, refName, refCulture
                            ).as_posix()
                        )
    return probedirs



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import json
import os
import pathlib

import click


# pylint: disable=too-few-public-methods
class SpecimenConfigParamType(click.Path):
    """
    A custom Click parameter type for handling configuration paths.
    This class extends `click.Path` to provide additional functionality for
    handling different types of configuration paths, including files, directories,
    and JSON configuration files.
    Attributes:
        name (str): The name of the parameter type, set to "config".
    Methods:
        convert(value, param, ctx):
            Converts the input value based on its prefix and returns the appropriate
            configuration data. Supports the following prefixes:
            - "file:" for file paths
            - "dir:" for directory paths
            - "config:" for JSON configuration files
            If no prefix is provided, it attempts to determine if the value is a file
            and loads it as JSON if possible. Otherwise, it treats the value as a
            directory path.
    """

    name = "specimen_config"

    @staticmethod
    def _get_param_type(filename: str):
        """Determines the type and properties of a parameter based on its filename prefix.
        Args:
            filename (str): The filename string to analyze, optionally with a type prefix
        Returns:
            tuple: A 3-tuple containing:
                - str: The parameter type ('FILE', 'DIR', 'CONFIG', or '' for default)
                - Path: The filepath as a Path object
                - Path or None: The install prefix path (for files), directory path (for dirs),
                  or None (for config or default)
        """

        if filename.startswith("file:"):
            filepath = pathlib.Path(filename[5:])
            return "FILE", filepath, filepath.parent
        if filename.startswith("dir:"):
            filepath = pathlib.Path(filename[4:])
            return "DIR", filepath, filepath
        if filename.startswith("config:"):
            filepath = pathlib.Path(filename[7:])
            return "CONFIG", filepath, None
        return "", pathlib.Path(filename), None

    def convert(self, value, param, ctx):
        # value received may already be the right type
        if isinstance(value, list):
            return value

        param_type, filepath, installprefix = self._get_param_type(value)

        # validate filepath exists and is readable
        if not filepath.exists():
            self.fail(f"{value!r} does not exist", param, ctx)
        if not os.access(filepath, os.R_OK):
            self.fail(f"{value!r} is not readable", param, ctx)

        # no explicit type given, use heuristics to determine correct type
        if not param_type:
            # if it's a file (that ends in .json) then treat it as a CONFIG file
            # if it's not a file then probably a directory (or something odd...)
            if filepath.is_file():
                if filepath.suffix.lower() == ".json":
                    param_type = "CONFIG"
                else:
                    param_type = "FILE"
                    installprefix = filepath.parent
            else:
                param_type = "DIR"
                installprefix = filepath

        if param_type in ("FILE", "DIR"):
            # avoid a relative directory of "./" as the install prefix
            if not installprefix.is_absolute() and len(installprefix.parts) == 0:
                installprefix = ""
            else:
                installprefix = installprefix.as_posix()
            # emulate a configuration file with the given path
            config = [{"extractPaths": [filepath.as_posix()], "installPrefix": installprefix}]
        elif param_type in ("CONFIG"):
            with click.open_file(filepath) as f:
                try:
                    config = json.load(f)
                except json.decoder.JSONDecodeError as err:
                    self.fail(
                        f"{filepath.as_posix()!r} config file contains invalid JSON", param, ctx
                    )

                for entry in config:
                    extract_path = entry["extractPaths"]
                    for pth in extract_path:
                        extract_path_convert = pathlib.Path(pth)
                        if not extract_path_convert.exists():
                            self.fail(f"invalid extract path in config file: {pth}", param, ctx)
        else:
            self.fail(f"{value!r} is not a valid specimen config type", param, ctx)

        return config



import json
import subprocess
import sys
import uuid
from datetime import datetime
from pathlib import Path

from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def run_cve_bin_tool(input_file_path, shaHash, output_dir):
    cvebin_file_name = f"{shaHash}_{input_file_path.stem}.json"
    output_file_path = output_dir / cvebin_file_name

    cdxvex_file_name = f"{shaHash}_{input_file_path.stem}.cdxvex"
    vex_output_path = output_dir / cdxvex_file_name

    try:
        command = [
            "cve-bin-tool",
            "--offline",
            "--input-file",
            str(input_file_path),
            "--output",
            str(output_file_path),
            "--format",
            "json",
            "--vex",
            str(vex_output_path),
        ]
        result = subprocess.run(command, capture_output=True, text=True, check=False)

        # Check the exit status
        if result.returncode in (0, 1):
            return output_file_path  # Return path to the generated JSON file
    except subprocess.CalledProcessError as e:
        logger.error(
            f"Error running CVE-bin-tool: {e}\nOutput: {e.output}\nError: {e.stderr}",
            file=sys.stderr,
        )
    return None


def convert_cve_to_openvex(json_output_path, shaHash, output_dir):
    openvex_file_name = f"{json_output_path.stem}.vex"
    openvex_output = output_dir / openvex_file_name

    # Open and read the .json file
    try:
        with open(json_output_path, "r") as file:
            cve_data = json.load(file)
    except json.JSONDecodeError as e:
        logger.error(f"Error reading JSON file: {e}")
        return
    except IOError as e:
        logger.error(f"IO error when reading {json_output_path}: {e}")

    openvex_template = {
        "@context": "https://openvex.dev/ns/v0.2.0",
        "@id": f"urn:uuid:{uuid.uuid4()}",
        "author": "Surfactant plugin cvebintool2vex",
        "timestamp": datetime.now().isoformat(),
        "version": 1,
        "tooling": "Surfactant (https://github.com/LLNL/Surfactant)",
        "statements": [],
    }

    for entry in cve_data:
        # Convert CVE data to OpenVEX format
        statement = {
            "vulnerability": {"name": entry["cve_number"]},
            "products": [
                {"@id": f"cpe:2.3:a:{entry['vendor']}:{entry['product']}:{entry['version']}:::::"}
            ],
            "status": "under_investigation",
            "source": entry["source"],
            "cvss_version": entry["cvss_version"],
            "cvss_vector": entry["cvss_vector"],
            "severity": entry["severity"],
        }
        openvex_template["statements"].append(statement)

    # Save the OpenVEX output to a new file
    try:
        with open(openvex_output, "w") as outfile:
            json.dump(openvex_template, outfile, indent=4)
    except IOError as e:
        logger.error(f"IO error when writing {openvex_output}: {e}")


def process_input(input_path, shaHash, output_dir=None):
    input_path = Path(input_path)
    if output_dir is None:
        output_dir = Path.cwd()

    if input_path.is_dir():
        for input_file in input_path.glob("*.*"):
            if input_file.suffix.lower() not in [".bin", ".exe", ".jar"]:
                continue
            process_file(input_file, shaHash, output_dir)
    elif input_path.is_file():
        process_file(input_path, shaHash, output_dir)
    else:
        logger.info(f"Error: {input_path} is neither a file nor a directory.")


def process_file(input_file, shaHash, output_directory):
    try:
        json_output_path = run_cve_bin_tool(input_file, shaHash, output_directory)
        if json_output_path:
            convert_cve_to_openvex(json_output_path, shaHash, output_directory)
    except subprocess.CalledProcessError as e:
        logger.error(f"Error running CVE-bin-tool: {e}")
    except json.JSONDecodeError as e:
        logger.error(f"JSON decoding error in {input_file}: {e}")
    except IOError as e:
        logger.error(f"I/O error processing {input_file}: {e}")

    # Check if the expected JSON file was created and proceed if it exists
    cvebin_file_name = f"{shaHash}_{input_file.stem}.json"
    jsonfile = output_directory / cvebin_file_name
    if not jsonfile.exists():
        logger.warning(f"Expected JSON file does not exist: {jsonfile}")


def delete_extra_files(*file_paths):
    for file_path in file_paths:
        try:
            if file_path.exists():
                file_path.unlink()
        except PermissionError as e:
            logger.error(f"Permission error deleting {file_path}: {e}")

        except OSError as e:
            logger.error(f"OS error deleting {file_path}: {e}")


@surfactant.plugin.hookimpl(specname="extract_file_info")
def cvebintool2vex(sbom: SBOM, software: Software, filename: str, filetype: str):
    """
    :param sbom(SBOM): The SBOM that the software entry/file is being added to. Can be used to add observations or analysis data.
    :param software(Software): The software entry associated with the file to extract information from.
    :param filename (str): The full path to the file to extract information from.
    :param filetype (str): File type information based on magic bytes.
    """
    # Only parsing executable files
    if filetype not in ["ELF", "PE"]:
        pass

    shaHash = str(software.sha256)
    filename = Path(filename)
    output_dir = Path.cwd()

    existing_json_path = output_dir / f"{shaHash}_additional_metadata.json"
    if existing_json_path.exists():
        with open(existing_json_path, "r") as file:
            data = json.load(file)
    else:
        data = {
            "sha256hash": shaHash,
            "filename": [filename.name],
            "openvex": [],
            "cyclonedx-vex": [],
            "cve-bin-tool": [],
        }

    # Assuming JSON, CDXVEX, and VEX files are processed here
    process_input(filename, shaHash, output_dir)
    # and you have the output files: .json, .cdxvex, .vex

    # Integrate .cdxvex and .vex file contents
    cdxvex_file_path = output_dir / f"{shaHash}_{filename.stem}.cdxvex"
    vex_file_path = output_dir / f"{shaHash}_{filename.stem}.vex"
    json_file_path = output_dir / f"{shaHash}_{filename.stem}.json"

    if cdxvex_file_path.exists() and vex_file_path.exists() and json_file_path.exists():
        # For .cdxvex and .vex files, if they contain JSON, parse them as such; otherwise, read as text
        try:
            with open(cdxvex_file_path, "r") as file:
                cdxvex_data = json.load(file)  # Assuming .cdxvex file is in JSON format
            data["cyclonedx-vex"].append(cdxvex_data)
        except json.JSONDecodeError:
            with open(cdxvex_file_path, "r") as file:
                cdxvex_data = file.read()  # Fallback if not JSON
            data["cyclonedx-vex"].append(cdxvex_data)

        try:
            with open(vex_file_path, "r") as file:
                vex_data = json.load(file)  # Assuming .vex file is in JSON format
            data["openvex"].append(vex_data)
        except json.JSONDecodeError:
            with open(vex_file_path, "r") as file:
                vex_data = file.read()  # Fallback if not JSON
            data["openvex"].append(vex_data)

        with open(json_file_path, "r") as file:
            json_data = json.load(file)
        data["cve-bin-tool"].append(json_data)

    # Attempt to save the updated data
    try:
        with open(existing_json_path, "w") as file:
            json.dump(data, file, indent=4)
            logger.info(f"Updated data saved to {existing_json_path}")
    except IOError as e:
        logger.error(f"IO error when writing {existing_json_path}: {e}")

    # Clean up extra files
    delete_extra_files(cdxvex_file_path, vex_file_path, json_file_path)


@surfactant.plugin.hookimpl
def update_db():
    # Example update logic
    try:
        subprocess.check_call(["cve-bin-tool", "--update", "now", "."])
        return "Database updated successfully."
    except subprocess.CalledProcessError as e:
        return f"Failed to update database: {e}"



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from .config import ContextEntry

try:
    from ._version import __version__, __version_tuple__
except ModuleNotFoundError:
    __version__ = ""
    __version_tuple__ = ()

__all__ = ["ContextEntry"]



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import uuid as uuid_module
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Set

from dataclasses_json import dataclass_json
from loguru import logger

from ._analysisdata import AnalysisData
from ._file import File
from ._hardware import Hardware
from ._observation import Observation
from ._provenance import SoftwareProvenance
from ._relationship import Relationship, StarRelationship
from ._software import Software, SoftwareComponent
from ._system import System

INTERNAL_FIELDS = {"software_lookup_by_sha256"}


@dataclass_json
@dataclass
class SBOM:
    # pylint: disable=R0902
    systems: List[System] = field(default_factory=list)
    hardware: List[Hardware] = field(default_factory=list)
    software: List[Software] = field(default_factory=list)
    relationships: Set[Relationship] = field(default_factory=set)
    analysisData: List[AnalysisData] = field(default_factory=list)
    observations: List[Observation] = field(default_factory=list)
    starRelationships: Set[StarRelationship] = field(default_factory=set)
    software_lookup_by_sha256: Dict = field(default_factory=dict)

    def __post_init__(self):
        self.__dataclass_fields__ = {
            k: v for k, v in self.__dataclass_fields__.items() if k not in INTERNAL_FIELDS
        }

    def add_relationship(self, rel: Relationship) -> None:
        self.relationships.add(rel)

    def create_relationship(self, xUUID: str, yUUID: str, relationship: str) -> Relationship:
        rel = Relationship(xUUID, yUUID, relationship)
        self.relationships.add(rel)
        return rel

    def find_relationship_object(self, relationship: Relationship) -> bool:
        return relationship in self.relationships

    def find_relationship(self, xUUID: str, yUUID: str, relationship: str) -> bool:
        return Relationship(xUUID, yUUID, relationship) in self.relationships

    def has_relationship(
        self,
        xUUID: Optional[str] = None,
        yUUID: Optional[str] = None,
        relationship: Optional[str] = None,
    ) -> bool:
        for rel in self.relationships:
            # We iterate until we find a relationship that meets all the conditions
            if xUUID and rel.xUUID != xUUID:
                continue
            if yUUID and rel.yUUID != yUUID:
                continue
            if relationship and rel.relationship.upper() != relationship.upper():
                continue
            return True
        return False

    def find_software(self, sha256: Optional[str]) -> Optional[Software]:
        if sha256 in self.software_lookup_by_sha256:
            return self.software_lookup_by_sha256[sha256]
        return None

    def add_software(self, sw: Software) -> None:
        if sw.sha256 is not None:
            self.software_lookup_by_sha256[sw.sha256] = sw
        self.software.append(sw)

    def add_software_entries(
        self, entries: Optional[List[Software]], parent_entry: Optional[Software] = None
    ):
        """Add software entries to the SBOM, merging into existing entries as needed.

        Args:
            entries (Optional[List[Software]]): A list of Software entries to add to the SBOM.
            parent_entry (Optional[Software]): An optional parent software entry to add "Contains" relationships to.
        """
        if not entries:
            return
        # if a software entry already exists with a matching file hash, augment the info in the existing entry
        for e in entries:
            existing_sw = self.find_software(e.sha256)
            if existing_sw and Software.check_for_hash_collision(existing_sw, e):
                logger.warning(
                    f"Hash collision between {existing_sw.name} and {e.name}; unexpected results may occur"
                )
            if not existing_sw:
                self.add_software(e)
            else:
                existing_uuid, entry_uuid = existing_sw.merge(e)
                # go through relationships and see if any need existing entries updated for the replaced uuid (e.g. merging SBOMs)
                for rel in self.relationships:
                    if rel.xUUID == entry_uuid:
                        rel.xUUID = existing_uuid
                    if rel.yUUID == entry_uuid:
                        rel.yUUID = existing_uuid
            # if a parent/container was specified for the file, add the new entry as a "Contains" relationship
            if parent_entry:
                parent_uuid = parent_entry.UUID
                child_uuid = existing_uuid if existing_sw else e.UUID
                # avoid duplicate relationships if the software entry already existed
                if not existing_sw or not self.find_relationship(
                    parent_uuid, child_uuid, "Contains"
                ):
                    self.create_relationship(parent_uuid, child_uuid, "Contains")
            # TODO a pass later on to check for and remove duplicate relationships should be added just in case

    # pylint: disable=too-many-arguments
    def create_software(
        self,
        *,  # all arguments are keyword-only
        name: Optional[str] = None,
        size: Optional[int] = None,
        sha1: Optional[str] = None,
        sha256: Optional[str] = None,
        md5: Optional[str] = None,
        fileName: Optional[List[str]] = None,
        installPath: Optional[List[str]] = None,
        containerPath: Optional[List[str]] = None,
        captureTime: Optional[int] = None,
        version: Optional[str] = None,
        vendor: Optional[List[str]] = None,
        description: Optional[str] = None,
        relationshipAssertion: Optional[str] = None,
        comments: Optional[str] = None,
        metadata: Optional[List[object]] = None,
        supplementaryFiles: Optional[List[File]] = None,
        provenance: Optional[List[SoftwareProvenance]] = None,
        recordedInstitution: Optional[str] = None,
        components: Optional[List[SoftwareComponent]] = None,
    ) -> Software:
        sw = Software(
            name=name,
            size=size,
            sha1=sha1,
            sha256=sha256,
            md5=md5,
            fileName=fileName,
            installPath=installPath,
            containerPath=containerPath,
            captureTime=captureTime,
            version=version,
            vendor=vendor,
            description=description,
            relationshipAssertion=relationshipAssertion,
            comments=comments,
            metadata=metadata,
            supplementaryFiles=supplementaryFiles,
            provenance=provenance,
            recordedInstitution=recordedInstitution,
            components=components,
        )
        self.software_lookup_by_sha256[sw.sha256] = sw
        self.software.append(sw)
        return sw

    def merge(self, sbom_m: SBOM):
        # merged/old to new UUID map
        uuid_updates = {}

        # merge systems entries
        if sbom_m.systems:
            for system in sbom_m.systems:
                # check for duplicate UUID/name, merge with existing entry
                if existing_system := self._find_systems_entry(uuid=system.UUID, name=system.name):
                    # merge system entries
                    u1, u2 = existing_system.merge(system)
                    logger.info(f"MERGE_DUPLICATE_SYS: uuid1={u1}, uuid2={u2}")
                    uuid_updates[u2] = u1
                else:
                    self.systems.append(system)

        # merge software entries
        if sbom_m.software:
            for sw in sbom_m.software:
                # NOTE: Do we want to pass in teh UUID here? What if we have two different UUIDs for the same file? Should hashes be required?
                if existing_sw := self._find_software_entry(
                    uuid=sw.UUID, sha256=sw.sha256, md5=sw.md5, sha1=sw.sha1
                ):
                    u1, u2 = existing_sw.merge(sw)
                    logger.info(f"MERGE DUPLICATE: uuid1={u1}, uuid2={u2}")
                    uuid_updates[u2] = u1
                else:
                    self.software.append(sw)

        # merge relationships
        if sbom_m.relationships:
            for rel in sbom_m.relationships:
                # rewrite UUIDs before doing the search
                if rel.xUUID in uuid_updates:
                    rel.xUUID = uuid_updates[rel.xUUID]
                if rel.yUUID in uuid_updates:
                    rel.yUUID = uuid_updates[rel.yUUID]
                if existing_rel := self._find_relationship_entry(
                    xUUID=rel.xUUID,
                    yUUID=rel.yUUID,
                    relationship=rel.relationship,
                ):
                    logger.info(f"DUPLICATE RELATIONSHIP: {existing_rel}")
                else:
                    self.relationships.add(rel)

        # rewrite container path UUIDs using rewrite map/list
        for sw in self.software:
            if sw.containerPath:
                for idx, path in enumerate(sw.containerPath):
                    u = path[:36]
                    # if container path starts with an invalid uuid4, sbom might not be valid
                    if self.is_valid_uuid4(u):
                        if u in uuid_updates:
                            updated_path = path.replace(u, uuid_updates[u], 1)
                            sw.containerPath[idx] = updated_path
                # remove duplicates
                sw.containerPath = [*set(sw.containerPath)]
        logger.info(f"UUID UPDATES: {uuid_updates}")
        # merge analysisData
        if sbom_m.analysisData:
            for analysisData in sbom_m.analysisData:
                self.analysisData.append(analysisData)
        # merge observations
        if sbom_m.observations:
            for observation in sbom_m.observations:
                self.observations.append(observation)
        # merge starRelationships
        if sbom_m.starRelationships:
            for star_rel in sbom_m.starRelationships:
                # rewrite UUIDs before doing the search
                if star_rel.xUUID in uuid_updates:
                    star_rel.xUUID = uuid_updates[star_rel.xUUID]
                if star_rel.yUUID in uuid_updates:
                    star_rel.yUUID = uuid_updates[star_rel.yUUID]
                if existing_star_rel := self._find_star_relationship_entry(
                    xUUID=star_rel.xUUID,
                    yUUID=star_rel.yUUID,
                    relationship=star_rel.relationship,
                ):
                    logger.info(f"DUPLICATE STAR RELATIONSHIP: {existing_star_rel}")
                else:
                    self.starRelationships.add(star_rel)

    def _find_systems_entry(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[System]:
        """Merge helper function to find and return
        the matching system entry in the provided sbom.

        Args:
            uuid (Optional[str]): The uuid of the desired system entry.
            name (Optional[str]): The name of the desired system entry.

        Returns:
            Optional[System]: The system found that matches the given criteria, otherwise None.
        """
        for system in self.systems:
            if uuid:
                if system.UUID != uuid:
                    continue
            if name:
                if system.name != name:
                    continue
            return system
        return None

    def _find_software_entry(
        self,
        uuid: Optional[str] = None,
        sha256: Optional[str] = None,
        md5: Optional[str] = None,
        sha1: Optional[str] = None,
    ) -> Optional[Software]:
        """Merge helper function to find and return
        the matching software entry in the provided sbom.

        Args:
            uuid (Optional[str]): The uuid of the desired software entry to match against if no hashes were provided.
            sha256 (Optional[str]): The sha256 of the desired software entry.
            md5 (Optional[str]): The md5 of the desired software entry.
            sha1 (Optional[str]): The sha1 of the desired software entry.

        Returns:
            Optional[Software]: The software entry found that matches the given criteria, otherwise None.
        """
        for sw in self.software:
            match = False
            # If we have hashes to check
            if sha256 or md5 or sha1:
                # Check if we have both sides of the comparison, then compare. At least one hash must match
                if sw.sha256 and sha256:
                    if sw.sha256 == sha256:
                        match = True
                if sw.md5 and md5:
                    if sw.md5 == md5:
                        match = True
                if sw.sha1 and sha1:
                    if sw.sha1 == sha1:
                        match = True
            # If no hashes to check, match by UUID
            else:
                if sw.UUID == uuid:
                    match = True
            if match:
                return sw
        return None

    def _find_relationship_entry(
        self,
        xUUID: Optional[str] = None,
        yUUID: Optional[str] = None,
        relationship: Optional[str] = None,
    ) -> Optional[Relationship]:
        """Merge helper function to find and return
        the matching relationship entry in the provided sbom.

        Args:
            xUUID (Optional[str]): The xUUID of the desired relationship entry.
            yUUID (Optional[str]): The yUUID of the desired relationship entry.
            relationship (Optional[str]): The relationship type of the desired relationship entry.

        Returns:
            Optional[Relationship]: The relationship entry found that matches the given criteria, otherwise None.
        """
        for rel in self.relationships:
            if xUUID:
                if rel.xUUID != xUUID:
                    continue
            if yUUID:
                if rel.yUUID != yUUID:
                    continue
            if relationship:
                if rel.relationship != relationship:
                    continue
            return rel
        return None

    def _find_star_relationship_entry(
        self,
        xUUID: Optional[str] = None,
        yUUID: Optional[str] = None,
        relationship: Optional[str] = None,
    ) -> Optional[StarRelationship]:
        """Merge helper function to find and return
        the matching star relationship entry in the provided sbom.

        Args:
            xUUID (Optional[str]): The xUUID of the desired relationship entry.
            yUUID (Optional[str]): The yUUID of the desired relationship entry.
            relationship (Optional[str]): The relationship type of the desired relationship entry.

        Returns:
            Optional[StarRelationship]: The star relationship found that matches the given criteria, otherwise None.
        """
        for rel in self.starRelationships:
            if xUUID:
                if rel.xUUID != xUUID:
                    continue
            if yUUID:
                if rel.yUUID != yUUID:
                    continue
            if relationship:
                if rel.relationship != relationship:
                    continue
            return rel
        return None

    def is_valid_uuid4(self, u: str) -> bool:
        """Merge helper function to check if a uuid is valid.

        Args:
            u (str):  The UUID to check.

        Returns:
            bool: True if the UUID is valid, otherwise False.
        """
        try:
            u_test = uuid_module.UUID(u, version=4)
        except ValueError:
            return False
        return str(u_test) == u



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from typing import Any, Dict

import olefile

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def supports_file(filetype) -> bool:
    return filetype == "OLE"


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    if not supports_file(filetype):
        return None
    return extract_ole_info(filename)


def extract_ole_info(filename: str) -> object:
    file_details: Dict[str, Any] = {}

    ole = olefile.OleFileIO(filename)
    md = ole.get_metadata()
    file_details["ole"] = {}

    # to check if an OLE is an MSI file, check the root storage object CLSID
    # {000c1084-0000-0000-c000-000000000046}	MSI
    # {000c1086-0000-0000-c000-000000000046}    Windows Installer Patch MSP
    # extensions are typically .msi and .msp for files with these two clsid's
    # less common would be a .msm (merge) with the same clsid as MSI
    # as well as .mst (transform) with a clsid of 000c1082
    if ole.root and hasattr(ole.root, "clsid"):
        file_details["ole"]["clsid"] = str(ole.root.clsid).lower()
        if file_details["ole"]["clsid"] == "000c1082-0000-0000-c000-000000000046":
            file_details["ole"]["clsid_type"] = "MST"
        if file_details["ole"]["clsid"] == "000c1084-0000-0000-c000-000000000046":
            file_details["ole"]["clsid_type"] = "MSI"  # or msm, depending on file extension
        if file_details["ole"]["clsid"] == "000c1086-0000-0000-c000-000000000046":
            file_details["ole"]["clsid_type"] = "MSP"

    for prop in md.SUMMARY_ATTRIBS:
        if value := getattr(md, prop, None):
            if isinstance(value, bytes):
                file_details["ole"][prop] = value.decode("unicode_escape")
            else:
                file_details["ole"][prop] = str(value)
    ole.close()
    return file_details



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import os
import pathlib

from surfactant.filetypeid.id_magic import identify_file_type
from surfactant.infoextractors.a_out_file import extract_file_info
from surfactant.sbomtypes import SBOM, Software

# Files are relative to tests/data/a_out_files
_file_to_machine_type = {
    "big_m68020.aout": "M68020",
    "big_netbsd_i386.aout": "NetBSD/i386",
    "big_netbsd_sparc.aout": "NetBSD/SPARC",
    "little_386.aout": "386",
    "little_unknown.aout": "Unknown",
}


def test_a_out_machine_type():
    base_path = pathlib.Path(__file__).parent.absolute()
    data_dir = os.path.join(base_path, "..", "data", "a_out_files")
    for file_name, machine_type in _file_to_machine_type.items():
        file_path = os.path.join(data_dir, file_name)
        file_type = identify_file_type(file_path)
        file_info = extract_file_info(SBOM(), Software(), file_path, file_type)
        assert file_info["aoutMachineType"] == machine_type



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import List, Optional

# pylint: disable=too-many-instance-attributes


@dataclass
class SystemProvenance:
    fieldName: str  # enum: captureStart, captureEnd, name, officialName, vendor, description
    capturedBy: Optional[str] = None
    captureTime: Optional[int] = None
    source: Optional[str] = None
    methodOfAcquisition: Optional[List[str]] = None


@dataclass
class HardwareProvenance:
    fieldName: str  # enum: name, captureTime, officialName, countryOfOrigin, countryOfOriginSource, quantity, description, vendor, identifiers, hardwareType, comments, metadata, packageType, boardLocation, recordedInstitution
    capturedBy: Optional[str] = None
    captureTime: Optional[int] = None
    source: Optional[str] = None
    methodOfAcquisition: Optional[List[str]] = None


@dataclass
class SoftwareComponentProvenance:
    fieldName: str  # enum: name, captureTime, version, vendor, description, comments, metadata, recordedInstitution
    capturedBy: Optional[str] = None
    captureTime: Optional[int] = None
    source: Optional[str] = None
    methodOfAcquisition: Optional[List[str]] = None


@dataclass
class SoftwareProvenance:
    fieldName: str  # enum: name, size, fileName, installPath, containerPath, captureTime, version, vendor, description, sha1, sha256, md5, relationshipAssertion, comments, metadata, recordedInstitution
    capturedBy: Optional[str] = None
    captureTime: Optional[int] = None
    source: Optional[str] = None
    methodOfAcquisition: Optional[List[str]] = None


@dataclass
class AnalysisDataProvenance:
    fieldName: (
        str  # enum: origin, testName, testVersion, specificEnvironment, linksToKnownVulnerabilities
    )
    capturedBy: Optional[str] = None
    captureTime: Optional[int] = None
    source: Optional[str] = None
    methodOfAcquisition: Optional[List[str]] = None


@dataclass
class ObservationProvenance:
    fieldName: str  # enum: flag, CWEClass, targetEnvironmentOrDevice, potentialEffectOrImpact, CVE, CVSS, analystInfo, discovery, toRecreate, mitigationSuggestions, confidenceLevel
    capturedBy: Optional[str] = None
    captureTime: Optional[int] = None
    source: Optional[str] = None
    methodOfAcquisition: Optional[List[str]] = None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

from surfactant.plugin.manager import get_plugin_manager
from surfactant.sbomtypes import SBOM, Relationship, Software

sbom = SBOM(
    software=[
        Software(
            UUID="application",
            fileName=["application.exe"],
            installPath=["C:\\application.exe"],
            metadata=[
                {
                    "peImport": ["library.dll"],
                }
            ],
        ),
        Software(
            UUID="library", fileName=["library.dll"], installPath=["C:\\library.dll"], metadata=[{}]
        ),
    ],
)


def test_same_directory():
    plugin = get_plugin_manager().get_plugin("surfactant.relationships.pe_relationship")
    app = sbom.software[0]
    md = app.metadata[0]
    assert plugin.establish_relationships(sbom, app, md) == [
        Relationship("application", "library", "Uses")
    ]



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import csv
import os
from collections.abc import Iterable
from typing import List, Optional

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software

default_fields = [
    "Path",
    "SHA1",
    "Supplier",
    "Product",
    "Version",
    "Description",
    "Copyright",
]


@surfactant.plugin.hookimpl
def write_sbom(sbom: SBOM, outfile) -> None:
    # plugin args could be handled here to change behavior
    fields = default_fields

    # match output format with pandas.DataFrame.to_csv
    # equivalent to `excel` dialect, other than lineterminator
    writer = csv.DictWriter(outfile, fieldnames=fields, lineterminator=os.linesep)
    writer.writeheader()
    if sbom.software:
        for sw in sbom.software:
            write_software_entry(writer, sw, fields)


@surfactant.plugin.hookimpl
def short_name() -> Optional[str]:
    return "csv"


def write_software_entry(writer: csv.DictWriter, software: Software, fields: List[str]):
    # last resort, use the fileName instead of an actual path to output csv entries
    pathkey = "fileName"
    if "Path" in fields:
        if software.installPath and isinstance(software.installPath, Iterable):
            # default to using "installPath"
            pathkey = "installPath"
        elif software.containerPath and isinstance(software.containerPath, Iterable):
            # use "containerPath" if it has entries but "installPath" does not
            pathkey = "containerPath"

    # an entry will be created for every entry with a valid path
    for p in getattr(software, pathkey):
        row = {}
        row["Path"] = p
        # if containerPath is being used, remove the UUID portion at the start
        if pathkey == "containerPath":
            row["Path"] = "".join(row["Path"].split("/")[1:])
        for f in fields:
            # Path already added to row info
            if f == "Path":
                continue
            # normalize some special field names to actual SBOM field names
            fld_norm = f
            if f in ("SHA1", "SHA256", "MD5", "Version", "Description"):
                fld_norm = str.lower(f)
            elif f == "Product":
                fld_norm = "name"
            elif f == "Supplier":
                fld_norm = "vendor"
            row[f] = get_software_field(software, fld_norm)
        writer.writerow(row)


def get_software_field(software, field):
    if hasattr(software, field):
        return getattr(software, field)
    # Copyright field currently only gets populated from Windows PE file metadata
    if field == "Copyright":
        if software.metadata and isinstance(software.metadata, Iterable):
            retval = []
            for entry in software.metadata:
                if "FileInfo" in entry and "LegalCopyright" in entry["FileInfo"]:
                    return entry["FileInfo"]["LegalCopyright"]
    return None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import datetime
import json
import os
import sys
import tempfile

import click
import flask

import surfactant.cmd.generate as gen
import surfactant.cmd.merge as merge_command

site = flask.Blueprint("site", __name__, url_prefix="/")


@site.route("/get_result/<path:result_id>")
def get_result(result_id):
    return flask.send_from_directory(f"{os.getcwd()}/results", result_id)


@site.route("/get_result_list")
def get_result_list():
    return json.dumps(
        sorted(
            [f for f in os.listdir("results") if os.path.isfile(os.path.join("results", f))],
            reverse=True,
        )
    )


@site.post("/generate")
def generate():
    # Hush some pylint things due to them not really being aplicable here
    # pylint: disable=broad-exception-caught, no-value-for-parameter
    try:
        req = flask.request.get_json()
        with tempfile.NamedTemporaryFile("w") as config_file:
            # Write the config to a temporary file
            config_input = []
            for inp in req["inputs"]:
                to_append = {}
                if inp["install_prefix"]:
                    to_append["installPrefix"] = inp["install_prefix"]
                if inp["archive"]:
                    to_append["archive"] = inp["archive"]
                to_append["extractPaths"] = inp["extract_paths"]
                config_input.append(to_append)
            json.dump(config_input, config_file)
            config_file.flush()

            output_file_name = (
                f"{datetime.datetime.now().strftime('%Y_%m_%d_%H_%M_%S_%f')}_generate.json"
            )

            with open(f"results/{output_file_name}", "w") as _:
                args = [config_file.name, f"results/{output_file_name}"]
                if req["input_sbom"]:
                    args.append(req["input_sbom"])
                if req["skip_gather"]:
                    args.append("--skip_gather")
                if req["skip_relationships"]:
                    args.append("--skip_relationships")
                if req["skip_install_path"]:
                    args.append("--skip_install_path")
                if req["recorded_institution"]:
                    args.append("--recorded_institution")
                    args.append(req["recorded_institution"])
                if req["input_format"]:
                    args.append("--input_format")
                    args.append(req["input_format"])
                if req["output_format"]:
                    args.append("--output_format")
                    args.append(req["output_format"])
                gen.sbom(args, standalone_mode=False)
                return {"error": False, "file_name": output_file_name}
    except Exception as e:
        return {"error": True, "error_desc": str(e)}


@site.post("/merge")
def merge():
    # pylint: disable=broad-exception-caught, no-value-for-parameter
    try:
        req = flask.request.get_json()
        output_file_name = f"{datetime.datetime.now().strftime('%Y_%m_%d_%H_%M_%S_%f')}_merge.json"

        with open(f"results/{output_file_name}", "w") as _:
            args = req["inputs"]
            args.append(f"results/{output_file_name}")
            if req["config_file"]:
                args.append("--config_file")
                args.append(req["config_file"])
            if req["input_format"]:
                args.append("--input_format")
                args.append(req["input_format"])
            if req["output_format"]:
                args.append("--output_format")
                args.append(req["output_format"])
            merge_command.merge_command(args, standalone_mode=False)
            return {"error": False, "file_name": output_file_name}
    except Exception as e:
        return {"error": True, "error_desc": str(e)}


@site.route("/")
def index():
    return flask.send_file(os.path.join(os.path.dirname(__file__), "../web-files/ui.html"))


@click.command("ui")
@click.argument("port", type=click.INT, required=False, default=8080)
def ui(port: int):
    """Start web UI server

    An optional PORT (default=8080) can be supplied as an argument.
    """
    os.makedirs("results", exist_ok=True)
    if not 0 <= port <= 65535:
        sys.exit("Port number must be between 0 and 65535")
    app = flask.Flask(__name__)
    app.register_blueprint(site)
    app.run(port=port)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import gzip
import json
import subprocess
import tempfile
from typing import Optional

from loguru import logger

import surfactant.plugin
from surfactant.configmanager import ConfigManager
from surfactant.sbomtypes import SBOM, Software


class DockerScoutManager:
    def __init__(self) -> None:
        # Initialize ConfigManager
        config_manager = ConfigManager()

        # Retrieve the configuration option
        enable_docker_scout = config_manager.get("docker", "enable_docker_scout", True)

        # Set disable_docker_scout based on the configuration
        self.disable_docker_scout = not enable_docker_scout
        self.docker_scout_installed = False

    def check_docker_scout_installed(self) -> None:
        """Check if Docker Scout is installed and update the state accordingly."""
        if self.disable_docker_scout:
            return  # Do nothing if Docker Scout is disabled by config

        try:
            result = subprocess.run(["docker", "scout"], capture_output=True, check=False)
            self.docker_scout_installed = result.returncode == 0
        except FileNotFoundError:
            self.docker_scout_installed = False

        self.disable_docker_scout = not self.docker_scout_installed
        if not self.docker_scout_installed:
            logger.warning(
                "Install Docker Scout to scan containers for additional information. "
                "You can also disable this check by running 'surfactant config docker.enable_docker_scout false'."
            )

    def run_docker_scout(self, filename: str) -> object:
        """Run Docker Scout on the given file and return the results."""
        if self.disable_docker_scout:
            return {}  # Do nothing if Docker Scout is disabled by config

        try:
            result = subprocess.run(
                ["docker", "scout", "sbom", "--format", "spdx", f"fs://{filename}"],
                capture_output=True,
                check=False,
            )
            if result.returncode != 0:
                logger.warning(f"Running Docker Scout on {filename} failed")
                return {}
            spdx_out = json.loads(result.stdout)
            return {"dockerSPDX": spdx_out}
        except json.JSONDecodeError as e:
            logger.error(f"Failed to parse Docker Scout output for {filename}: {e}")
            return {}


# Initialize DockerScoutManager to check installation status
dsManager = DockerScoutManager()


def supports_file(filetype: str) -> bool:
    """Check if the file type is supported."""
    return filetype in ("DOCKER_TAR", "DOCKER_GZIP")


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: str) -> object:
    """Extract file information using Docker Scout if supported."""
    if dsManager.disable_docker_scout or not supports_file(filetype):
        return None
    return extract_docker_info(filetype, filename)


def extract_docker_info(filetype: str, filename: str) -> object:
    """Extract Docker information based on file type."""
    if filetype == "DOCKER_GZIP":
        with open(filename, "rb") as gzip_in:
            gzip_data = gzip_in.read()
        with tempfile.NamedTemporaryFile() as gzip_out:
            gzip_out.write(gzip.decompress(gzip_data))
            gzip_out.flush()  # Ensure data is written before reading
            return dsManager.run_docker_scout(gzip_out.name)
    return dsManager.run_docker_scout(filename)


@surfactant.plugin.hookimpl
def init_hook(command_name: Optional[str] = None) -> None:
    if command_name != "update-db" and not dsManager.disable_docker_scout:
        dsManager.check_docker_scout_installed()



import os
import platform
from pathlib import Path
from threading import Lock
from typing import Any, Dict, Optional, Union

import tomlkit


class ConfigManager:
    """A configuration manager for handling settings stored in a configuration file. The
    configuration manager internally caches a copy of the loaded configuration file, so
    external changes won't affect the setting value while a program is running.

    Attributes:
        app_name (str): The name of the application. (Default: 'surfactant')
        config_dir (Optional[Path]): The directory where the configuration file is stored.
        config (tomlkit.document): The configuration document loaded by tomlkit. Preserves formatting and comments.
        config_file_path (Path): The path to the configuration file.
    """

    _initialized: bool = False
    _instances: Dict[str, "ConfigManager"] = {}
    _lock = Lock()

    def __new__(
        cls, app_name: str = "surfactant", config_dir: Optional[Union[str, Path]] = None
    ) -> "ConfigManager":
        """Manage singleton configuration manager for each unique application name.

        Args:
            app_name (str): The name of the application. (Default: 'surfactant')
            config_dir (Optional[Union[str, Path]]): The directory where the application configuration is stored.

        Returns:
            ConfigManager: The singleton instance of the configuration manager for the given application name.
        """
        with cls._lock:
            if app_name not in cls._instances:
                instance = super(ConfigManager, cls).__new__(cls)
                instance._initialized = False
                cls._instances[app_name] = instance
            return cls._instances[app_name]

    def __init__(
        self, app_name: str = "surfactant", config_dir: Optional[Union[str, Path]] = None
    ) -> None:
        """Initializes the configuration manager.

        Args:
            app_name (str): The name of the application. (Default: 'surfactant')
            config_dir (Optional[Union[str, Path]]): The directory where the application configuration is stored.
        """
        if self._initialized:
            return
        self._initialized = True

        self.app_name = app_name
        self.config_dir = Path(config_dir) / app_name if config_dir else None
        self.config = tomlkit.document()
        self.config_file_path = self._get_config_file_path()
        self._load_config()

    def _get_config_file_path(self) -> Path:
        """Determines the path to the configuration file.

        Returns:
            Path: The path to the configuration file.
        """
        if self.config_dir:
            config_dir = Path(self.config_dir)
        else:
            if platform.system() == "Windows":
                config_dir = Path(os.getenv("APPDATA", str(Path("~\\AppData\\Roaming"))))
            else:
                config_dir = Path(os.getenv("XDG_CONFIG_HOME", str(Path("~/.config"))))
            config_dir = config_dir / self.app_name / "config.toml"
        return config_dir.expanduser()

    def _load_config(self) -> None:
        """Loads the configuration from the configuration file."""
        if self.config_file_path.exists():
            with open(self.config_file_path, "r") as configfile:
                self.config = tomlkit.parse(configfile.read())

    def get(self, section: str, option: str, fallback: Optional[Any] = None) -> Any:
        """Gets a configuration value.

        Args:
            section (str): The section within the configuration file.
            option (str): The option within the section.
            fallback (Optional[Any]): The fallback value if the option is not found.

        Returns:
            Any: The configuration value or the fallback value.
        """
        return self.config.get(section, {}).get(option, fallback)

    def set(self, section: str, option: str, value: Any) -> None:
        """Sets a configuration value.

        Args:
            section (str): The section  within the configuration file.
            option (str): The option within the section.
            value (Any): The value to set.
        """
        if section not in self.config:
            self.config[section] = tomlkit.table()
        self.config[section][option] = value
        self._save_config()

    def _save_config(self) -> None:
        """Saves the configuration to the configuration file."""
        if not self.config_file_path.exists():
            self.config_file_path.parent.mkdir(parents=True, exist_ok=True)
        with open(self.config_file_path, "w") as configfile:
            configfile.write(tomlkit.dumps(self.config))

    def __getitem__(self, key: str) -> Any:
        """Enables dictionary-like syntax for accessing configuration settings.
        NOTE: Remember to check that the value returned is not 'None' before
        trying to access nested keys.

        Args:
            key (str): The key for accessing a TOML value or table.

        Returns:
            Any: The configuration value or 'NoneType' if the key doesn't exist.
        """
        if key not in self.config:
            return None
        return self.config[key]

    @classmethod
    def delete_instance(cls, app_name: str) -> None:
        """Deletes the singleton instance for the given application name.

        Args:
            app_name (str): The name of the application.
        """
        with cls._lock:
            if app_name in cls._instances:
                del cls._instances[app_name]

    def get_data_dir_path(self) -> Path:
        """Determines the path to the data directory, for storing things such as databases.

        Returns:
            Path: The path to the data directory.
        """
        if platform.system() == "Windows":
            data_dir = Path(os.getenv("LOCALAPPDATA", str(Path("~\\AppData\\Local"))))
        else:
            data_dir = Path(os.getenv("XDG_DATA_HOME", str(Path("~/.local/share"))))
        data_dir = data_dir / self.app_name
        return data_dir.expanduser()



from typing import Any, List, Optional

import click

from surfactant.configmanager import ConfigManager


@click.command("config")
@click.argument("key", required=True)
@click.argument("values", nargs=-1)
def config(key: str, values: Optional[List[str]]):
    """Get or set a configuration value.

    If only KEY is provided, the current value is displayed.
    If both KEY and one or more VALUES are provided, the configuration value is set.
    KEY should be in the format 'section.option'.
    """
    config_manager = ConfigManager()

    if not values:
        # Get the configuration value
        try:
            section, option = key.split(".", 1)
        except ValueError as err:
            raise SystemExit("Invalid KEY given. Is it in the format 'section.option'?") from err
        result = config_manager.get(section, option)
        if result is None:
            click.echo(f"Configuration '{key}' not found.")
        else:
            click.echo(f"{key} = {result}")
    else:
        # Set the configuration value
        # Convert 'true' and 'false' strings to boolean
        converted_values: List[Any] = []
        for value in values:
            if value.lower() == "true":
                converted_values.append(True)
            elif value.lower() == "false":
                converted_values.append(False)
            else:
                converted_values.append(value)

        # If there's only one value, store it as a single value, otherwise store as a list
        final_value = converted_values[0] if len(converted_values) == 1 else converted_values

        try:
            section, option = key.split(".", 1)
        except ValueError as err:
            raise SystemExit("Invalid KEY given. Is it in the format 'section.option'?") from err
        config_manager.set(section, option, final_value)
        click.echo(f"Configuration '{key}' set to '{final_value}'.")



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

import json
import os
import pathlib
import random
import string

import pytest

from surfactant.cmd.merge import merge
from surfactant.plugin.manager import get_plugin_manager
from surfactant.sbomtypes import SBOM, Relationship

# Generate Sample SBOMs
sbom1_json = """{
  "software": [
    {
      "UUID": "dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
      "fileName": [
        "helics.tar.gz"
      ]
    },
    {
      "UUID": "08526f02-8f08-485d-bfd1-ea16ce964fd2",
      "fileName": [
        "helics_binary"
      ],
      "installPath": [
        "/bin/helics_binary"
      ],
      "containerPath": [
        "dd6f7f6b-7c31-4a4a-afef-14678b9942bf/bin/helics_binary"
      ]
    },
    {
      "UUID": "a5db7e12-fe3d-490e-90b8-98a8bfaace09",
      "fileName": [
        "lib1.so"
      ],
      "installPath": [
        "/lib64/lib1.so"
      ],
      "containerPath": [
        "dd6f7f6b-7c31-4a4a-afef-14678b9942bf/lib64/lib1.so"
      ]
    }
  ],
  "relationships": [
    {
      "xUUID": "08526f02-8f08-485d-bfd1-ea16ce964fd2",
      "yUUID": "a5db7e12-fe3d-490e-90b8-98a8bfaace09",
      "relationship": "Uses"
    },
    {
      "xUUID": "dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
      "yUUID": "08526f02-8f08-485d-bfd1-ea16ce964fd2",
      "relationship": "Contains"
    },
    {
      "xUUID": "dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
      "yUUID": "a5db7e12-fe3d-490e-90b8-98a8bfaace09",
      "relationship": "Contains"
    }
  ]
}"""


def get_sbom1():
    return SBOM.from_json(sbom1_json)


sbom2_json = """{
        "software": [
            {
            "UUID": "625a07da-7eed-47b9-a0fa-47dcbf76574a",
            "name": null,
            "size": null,
            "fileName": [
                "helics_plugin.tar.gz"
            ]
            },
            {
            "UUID": "820d3ddc-14d7-4ce5-833c-beede8725366",
            "fileName": [
                "helics_plugin"
            ],
            "installPath": [
                "/bin/helics_plugin"
            ],
            "containerPath": [
                "625a07da-7eed-47b9-a0fa-47dcbf76574a/bin/helics_plugin"
            ]
            },
            {
            "UUID": "df81b6a7-f9df-42f1-85ee-86a8865fa5f1",
            "fileName": [
                "lib_plugin.so"
            ],
            "installPath": [
                "/lib64/lib_plugin.so"
            ],
            "containerPath": [
                "625a07da-7eed-47b9-a0fa-47dcbf76574a/lib64/lib_plugin.so"
            ]
            }
        ],
        "relationships": [
            {
            "xUUID": "820d3ddc-14d7-4ce5-833c-beede8725366",
            "yUUID": "df81b6a7-f9df-42f1-85ee-86a8865fa5f1",
            "relationship": "Uses"
            },
            {
            "xUUID": "625a07da-7eed-47b9-a0fa-47dcbf76574a",
            "yUUID": "820d3ddc-14d7-4ce5-833c-beede8725366",
            "relationship": "Contains"
            },
            {
            "xUUID": "625a07da-7eed-47b9-a0fa-47dcbf76574a",
            "yUUID": "df81b6a7-f9df-42f1-85ee-86a8865fa5f1",
            "relationship": "Contains"
            }
        ]
        }"""


def get_sbom2():
    return SBOM.from_json(sbom2_json)


def get_config():
    return {
        "system": {
            "UUID": "6a0ee431-842f-4963-8867-ef0ef6998003",
            "name": "",
            "vendor": None,
            "captureStart": 1689186121,
            "captureEnd": 1689186146,
        }
    }


def get_sbom3():
    with open(
        pathlib.Path(__file__).parent / "../data/sample_sboms/helics_binaries_sbom.json",
        "r",
    ) as f:
        return SBOM.from_json(f.read())


def get_sbom4():
    with open(
        pathlib.Path(__file__).parent / "../data/sample_sboms/helics_libs_sbom.json", "r"
    ) as f:
        return SBOM.from_json(f.read())


# Test Functions
def test_simple_merge_method():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    merged_sbom = sbom1
    merged_sbom.merge(sbom2)
    softwares = sbom1.software
    softwares.extend(sbom2.software)
    assert merged_sbom.software.sort(key=lambda x: x.UUID) == softwares.sort(key=lambda x: x.UUID)
    relations = sbom1.relationships
    relations.update(sbom2.relationships)
    assert sorted(merged_sbom.relationships, key=lambda x: x.xUUID) == sorted(
        relations, key=lambda x: x.xUUID
    )


@pytest.mark.skip(reason="No way of validating this test yet")
def test_merge_with_circular_dependency():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    circular_dependency_sbom = sbom1
    circular_dependency_sbom.relationships.add(
        Relationship(
            xUUID="a5db7e12-fe3d-490e-90b8-98a8bfaace09",
            yUUID="dd6f7f6b-7c31-4a4a-afef-14678b9942bf",
            relationship="Contains",
        )
    )

    outfile_name = generate_filename("test_merge_with_circular_dependency")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [circular_dependency_sbom, sbom2]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, get_config(), output_writer)
    # TODO add validation checks here
    os.remove(os.path.abspath(outfile_name))


@pytest.mark.skip(reason="No way of properly validating this test yet")
def test_cmdline_merge():
    sbom3 = get_sbom3()
    sbom4 = get_sbom4()
    # Test simple merge of two sboms
    outfile_name = generate_filename("test_cmdline_merge")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    config_file = get_config()
    input_sboms = [sbom3, sbom4]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer)

    # TODO add validation checks here
    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    with open(pathlib.Path(__file__).parent / "../data/sample_sboms/helics_sbom.json", "r") as j:
        ground_truth_sbom = json.loads(j.read())
    os.remove(os.path.abspath(outfile_name))


def test_merge_with_add_system_true():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    outfile_name = generate_filename("test_merge_with_add_system_true")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    config_file["system"]["UUID"] = "6a0ee431-842f-4963-8867-ef0ef6998003"
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=True)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert generated_sbom["systems"]
    assert generated_sbom["systems"][0]["UUID"] == config_file["system"]["UUID"]

    os.remove(os.path.abspath(outfile_name))


def test_merge_with_add_system_false():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    outfile_name = generate_filename("test_merge_with_add_system_false")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    config_file["system"]["UUID"] = "6a0ee431-842f-4963-8867-ef0ef6998003"
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=False)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert not generated_sbom["systems"]

    os.remove(os.path.abspath(outfile_name))


def test_merge_with_custom_system_relationship():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    outfile_name = generate_filename("test_merge_with_custom_system_relationship")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    config_file["systemRelationship"] = "DependsOn"
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=True)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    for relationship in generated_sbom["relationships"]:
        if relationship["xUUID"] == config_file["system"]["UUID"]:
            assert relationship["relationship"] == "DependsOn"

    os.remove(os.path.abspath(outfile_name))


def test_merge_with_specified_system_uuid():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    outfile_name = generate_filename("test_merge_with_specified_system_uuid")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    system_uuid = "123e4567-e89b-12d3-a456-426614174000"
    with open(outfile_name, "w") as sbom_outfile:
        merge(
            input_sboms,
            sbom_outfile,
            config_file,
            output_writer,
            add_system=True,
            system_uuid=system_uuid,
        )

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert any(system["UUID"] == system_uuid for system in generated_sbom["systems"])

    os.remove(os.path.abspath(outfile_name))


def test_prevent_orphaned_system_uuid():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    outfile_name = generate_filename("test_prevent_orphaned_system_uuid")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    # Get rid of the system UUID field from the config file
    # This will make it try to generate a random UUID, but won't add it since add_system is False
    del config_file["system"]["UUID"]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=False)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert not generated_sbom["systems"]

    os.remove(os.path.abspath(outfile_name))


def test_add_random_system_uuid():
    sbom1 = get_sbom1()
    sbom2 = get_sbom2()
    outfile_name = generate_filename("test_prevent_orphaned_system_uuid")
    pm = get_plugin_manager()
    output_writer = pm.get_plugin("surfactant.output.cytrics_writer")
    input_sboms = [sbom1, sbom2]
    config_file = get_config()
    # Get rid of the system UUID field from the config file
    # This will make it try to generate a random UUID, but won't add it since add_system is False
    original_config_system_UUID = config_file["system"]["UUID"]
    del config_file["system"]["UUID"]
    with open(outfile_name, "w") as sbom_outfile:
        merge(input_sboms, sbom_outfile, config_file, output_writer, add_system=True)

    with open(outfile_name, "r") as j:
        generated_sbom = json.loads(j.read())
    assert generated_sbom["systems"]
    # Check that the UUID of the generated system is actually random
    assert generated_sbom["systems"][0]["UUID"] != original_config_system_UUID

    os.remove(os.path.abspath(outfile_name))


def generate_filename(name, ext=".json"):
    res = "".join(random.choices(string.ascii_uppercase + string.digits, k=7))
    return str(name + "_" + res + ext)



# Copyright 2024 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT

# For compatibility with old versions of tools
from setuptools import setup

setup()
